import abc

import numpy as np
from scipy.optimize import brute
from scipy.special import (
    log_softmax,
    softmax,
)


class ClassificationTemperatureScaling(abc.ABC):
    def __init__(self):
        """
        A temperature scaling class for classification. It scales the logits with a shared learnable parameters.
        """
        self._temperature = None

    def fit(self, probs: np.ndarray, targets: np.ndarray):
        """
        Fit che temperature.

        Parameters
        ----------
        probs: np.ndarray
            A two-dimensional array of probabilities, for each input and class.
        targets: np.ndarray
            A one-dimensional array of integer target variables.
        """
        self._check_probs(probs)
        self._check_targets(targets)
        log_probs = np.log(probs)
        one_hot_targets = np.eye(probs.shape[1])[targets.reshape(-1)]

        def temp_scaling_fn(tau):
            log_temp_probs = log_softmax(log_probs / tau, axis=1)
            return -np.sum(one_hot_targets * log_temp_probs)

        self._temperature = brute(temp_scaling_fn, ranges=[(1e-6, 10)], Ns=1000)[0]

    def predict_proba(self, probs: np.ndarray) -> np.ndarray:
        """
        Predict the scaled probabilities.

        Parameters
        ----------
        probs: np.ndarray
            A two-dimensional array of probabilities, for each input and class.

        Returns
        -------
        np.ndarray
            The predicted probabilities for each input and class.
        """
        self._check_probs(probs)
        return softmax(np.log(probs) / self._temperature, axis=1)

    def predict(self, probs: np.ndarray) -> np.ndarray:
        """
        Predict the target variable with the largest probability.

        Parameters
        ----------
        probs: np.ndarray
            A two-dimensional array of probabilities, for each input and class.

        Returns
        -------
        np.ndarray
            The predicted target variables for each input.
        """
        self._check_probs(probs)
        return np.argmax(probs, axis=1)

    @property
    def temperature(self):
        return self._temperature

    @staticmethod
    def _check_probs(probs: np.ndarray):
        if probs.ndim != 2:
            raise ValueError("The array of probabilities must be two-dimensional.")

    @staticmethod
    def _check_targets(targets: np.ndarray):
        if targets.ndim != 1:
            raise ValueError("The array of targets must be one-dimensional.")
        if targets.dtype != int:
            raise ValueError("Each element in the array of targets must be an integer.")



from fortuna.conformal.classification.adaptive_conformal_classifier import (
    AdaptiveConformalClassifier,
)
from fortuna.conformal.classification.adaptive_prediction import (
    AdaptivePredictionConformalClassifier,
    CVPlusAdaptivePredictionConformalClassifier,
)
from fortuna.conformal.classification.maxcovfixprec_binary_classfication import (
    MaxCoverageFixedPrecisionBinaryClassificationCalibrator,
)
from fortuna.conformal.classification.simple_prediction import (
    CVPlusSimplePredictionConformalClassifier,
    SimplePredictionConformalClassifier,
)
from fortuna.conformal.multivalid.iterative.classification.binary_multicalibrator import (
    BinaryClassificationMulticalibrator,
)
from fortuna.conformal.multivalid.iterative.classification.top_label_multicalibrator import (
    TopLabelMulticalibrator,
)
from fortuna.conformal.multivalid.iterative.multicalibrator import Multicalibrator
from fortuna.conformal.multivalid.iterative.regression.batch_mvp import (
    BatchMVPConformalClassifier,
)
from fortuna.conformal.multivalid.one_shot.classification.binary_multicalibrator import (
    OneShotBinaryClassificationMulticalibrator,
)
from fortuna.conformal.multivalid.one_shot.classification.top_label_multicalibrator import (
    OneShotTopLabelMulticalibrator,
)
from fortuna.conformal.multivalid.one_shot.multicalibrator import OneShotMulticalibrator
from fortuna.conformal.regression.adaptive_conformal_regressor import (
    AdaptiveConformalRegressor,
)
from fortuna.conformal.regression.batch_mvp import BatchMVPConformalRegressor
from fortuna.conformal.regression.cvplus import CVPlusConformalRegressor
from fortuna.conformal.regression.enbpi import EnbPI
from fortuna.conformal.regression.jackknife_minmax import (
    JackknifeMinmaxConformalRegressor,
)
from fortuna.conformal.regression.jackknifeplus import JackknifePlusConformalRegressor
from fortuna.conformal.regression.onedim_uncertainty import (
    OneDimensionalUncertaintyConformalRegressor,
)
from fortuna.conformal.regression.quantile import QuantileConformalRegressor



import logging
import os
from typing import (
    Dict,
    Optional,
)

from flax.training import checkpoints
from flax.training.early_stopping import EarlyStopping

from fortuna.training.name_to_train_state import NameToTrainState
from fortuna.training.train_state import TrainState
from fortuna.typing import (
    OptaxOptimizer,
    Path,
)

logger = logging.getLogger(__name__)


class WithCheckpointingMixin:
    def __init__(
        self,
        **kwargs,
    ):
        """
        Mixin class for all trainers that need checkpointing capabilities. This is a wrapper around functions in
        `flax.training.checkpoints.*`.
        """
        super(WithCheckpointingMixin, self).__init__(**kwargs)

    def save_checkpoint(
        self,
        state: TrainState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        if save_checkpoint_dir:
            save_ckpt_fn = lambda state: checkpoints.save_checkpoint(
                ckpt_dir=str(save_checkpoint_dir),
                target=state,
                step=state.step,
                prefix=prefix,
                keep=keep,
                overwrite=force_save,
            )
            if (
                hasattr(state, "grad_accumulated")
                and state.grad_accumulated is not None
            ):
                # do not save grad accumulated in the ckpt
                state = state.replace(grad_accumulated=None)
                save_ckpt_fn(state)
            else:
                save_ckpt_fn(state)

    def restore_checkpoint(
        self,
        restore_checkpoint_path: Path,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        name_to_train_state: NameToTrainState = NameToTrainState,
        **kwargs,
    ) -> TrainState:
        if not os.path.isdir(restore_checkpoint_path) and not os.path.isfile(
            restore_checkpoint_path
        ):
            raise ValueError(
                f"`restore_checkpoint_path={restore_checkpoint_path}` was not found."
            )
        d = checkpoints.restore_checkpoint(
            ckpt_dir=str(restore_checkpoint_path),
            target=None,
            step=None,
            prefix=prefix,
            parallel=True,
        )
        if d is None:
            raise ValueError(
                f"No checkpoint was found in `restore_checkpoint_path={restore_checkpoint_path}`."
            )
        name = "".join([chr(n) for n in d["encoded_name"].values()])
        return name_to_train_state[name].value.init_from_dict(d, optimizer, **kwargs)

    def get_path_latest_checkpoint(
        self, checkpoint_dir: Path, prefix: str = "checkpoint_"
    ) -> Optional[str]:
        return checkpoints.latest_checkpoint(ckpt_dir=checkpoint_dir, prefix=prefix)


class WithEarlyStoppingMixin:
    def __init__(
        self,
        *,
        early_stopping_monitor: str = "val_loss",
        early_stopping_min_delta: float = 0.0,
        early_stopping_patience: Optional[int] = 0,
        early_stopping_mode: str = "min",
        early_stopping_verbose: bool = True,
        **kwargs,
    ):
        super(WithEarlyStoppingMixin, self).__init__(**kwargs)
        self.early_stopping_monitor = early_stopping_monitor
        self.early_stopping_mode = early_stopping_mode
        self.early_stopping_patience = early_stopping_patience

        if early_stopping_patience is None or early_stopping_patience <= 0:
            if early_stopping_verbose:
                logging.info(
                    f"Early stopping not enabled. Set `early_stopping_patience>=0` to enable it."
                )
        elif self.early_stopping_mode is None or self.early_stopping_mode not in (
            "min",
            "max",
        ):
            if early_stopping_verbose:
                logging.warning(
                    f"`early_stopping_mode={early_stopping_mode}` is not a valid. Early stopping will be disabled."
                )
        else:
            self._early_stopping = EarlyStopping(
                min_delta=early_stopping_min_delta, patience=early_stopping_patience
            )
            if early_stopping_verbose:
                logging.info(
                    "If validation data are provided, early stopping will be enabled."
                )

    @property
    def is_early_stopping_active(self) -> bool:
        return not (
            (self.early_stopping_patience is None or self.early_stopping_patience <= 0)
            or (
                self.early_stopping_mode is None
                or self.early_stopping_mode not in ("min", "max")
            )
        )

    def early_stopping_update(
        self, validation_metrics: Dict[str, float]
    ) -> Optional[bool]:
        improved = None
        if self.is_early_stopping_active:
            early_stopping_monitor = validation_metrics[self.early_stopping_monitor]
            if self.early_stopping_mode == "max":
                early_stopping_monitor = -early_stopping_monitor
            improved, self._early_stopping = self._early_stopping.update(
                early_stopping_monitor
            )
        return improved


class InputValidatorMixin:
    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            raise AttributeError("Cannot recognize inputs arguments: {}".format(args))
        if len(kwargs) > 0:
            raise AttributeError(
                "{} are not valid input arguments.".format(list(kwargs.keys()))
            )
        super(InputValidatorMixin, self).__init__(*args, **kwargs)



import numpy as np

from fortuna.calibration.binary_classification.temp_scaling.base import (
    BaseBinaryClassificationTemperatureScaling,
)


class BiasBinaryClassificationTemperatureScaling(
    BaseBinaryClassificationTemperatureScaling
):
    """
    A temperature scaling class for binary classification.
    It scales the probability that the target variables is positive with a single learnable parameters.
    The method minimizes the expected bias.
    """

    def fit(self, probs: np.ndarray, targets: np.ndarray):
        self._check_probs(probs)
        self._check_targets(targets)
        self._temperature = np.mean(probs) / np.mean(targets)



from typing import Optional

from jax import vmap
import jax.numpy as jnp

from fortuna.conformal.multivalid.mixins.multicalibrator import MulticalibratorMixin
from fortuna.typing import Array


class TopLabelMulticalibratorMixin(MulticalibratorMixin):
    def __init__(self, n_classes: int, seed: int = 0):
        """
        A multicalibration method that provides multivalid coverage guarantees. See Algorithm 15 in `Aaron Roth's notes
        <https://www.cis.upenn.edu/~aaroth/uncertainty-notes.pdf>`_.

        Parameters
        ----------
        n_classes: int
            Number of classes.
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)
        self._patch_list = []
        self.n_classes = n_classes

    def mean_squared_error(self, probs: Array, targets: Array) -> Array:
        return super().mean_squared_error(
            values=probs, scores=self._get_scores(targets)
        )

    def _get_scores(self, targets: Array) -> Array:
        self._check_targets(targets)
        scores = []
        for i in range(self.n_classes):
            scores.append(targets == i)
        return jnp.stack(scores, axis=1)

    @staticmethod
    def _check_targets(targets: Array):
        if targets.ndim != 1:
            raise ValueError("`targets` must be a 1-dimensional array of integers.")
        if targets.dtype not in ["int32", "int64"]:
            raise ValueError("All elements in `targets` must be integers")

    @staticmethod
    def _check_scores(scores: Array):
        pass

    @staticmethod
    def _maybe_check_values(
        values: Optional[Array], test_values: Optional[Array] = None
    ):
        if values is not None:
            if values.ndim != 2:
                raise ValueError(
                    "`probs` must be a 2-dimensional array representing the probabilities for each input "
                    "and each class."
                )
            if jnp.any(values < 0) or jnp.any(values > 1):
                raise ValueError("All elements in `values` must be within [0, 1].")
        if test_values is not None:
            if test_values.ndim != 2:
                raise ValueError(
                    "`test_probs` must be a 2-dimensional array representing the probabilities for each "
                    "input and each class."
                )
            if jnp.any(test_values < 0) or jnp.any(test_values > 1):
                raise ValueError("All elements in `test_values` must be within [0, 1].")

    @staticmethod
    def _maybe_normalize(values: Array):
        sum_values = values.sum(1, keepdims=True)
        cond = jnp.isnan(sum_values) * (sum_values == 0.0)
        return values * cond + (values / sum_values) * (1 - cond)



from typing import Optional

import flax.linen as nn
from flax.linen.initializers import Initializer
import jax.numpy as jnp

from fortuna.typing import Array


class ConstantModel(nn.Module):
    r"""
    A constant model, that is :math:`f(\theta, x) = \theta`.

    Parameters
    ----------
    output_dim: int
        The output model dimension.
    initializer_fun: Optional[Initializer]
        Function to initialize the model parameters.
        This must be one of the available options in :code:`flax.linen.initializers`.
    """
    output_dim: int
    initializer_fun: Optional[Initializer] = nn.initializers.zeros

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        constant = self.param("constant", self.initializer_fun, (self.output_dim,))
        return jnp.broadcast_to(constant, shape=(x.shape[0], self.output_dim))



from typing import Callable

import jax.numpy as jnp
import numpy as np

from fortuna.typing import Array

StepSchedule = Callable[[Array], Array]


def constant_schedule(init_step_size: float) -> StepSchedule:
    """Create a constant step schedule.

    Parameters
    ----------
    init_step_size: float
        The step size.

    Returns
    -------
    schedule_fn: StepSchedule
    """
    if not init_step_size >= 0:
        raise ValueError("`init_step_size` should be >= 0.")

    def schedule(_step: Array):
        return init_step_size

    return schedule


def cosine_schedule(init_step_size: float, total_steps: int) -> StepSchedule:
    """Create a cosine step schedule.

    Parameters
    ----------
    init_step_size: float
        The initial step size.
    total_steps: int
        The cycle length, in steps.

    Returns
    -------
    schedule_fn: StepSchedule
    """
    if not init_step_size >= 0:
        raise ValueError("`init_step_size` should be >= 0.")
    if not total_steps > 0:
        raise ValueError("`total_steps` should be > 0.")

    def schedule(step: Array):
        t = step / total_steps
        return 0.5 * init_step_size * (1 + jnp.cos(t * np.pi))

    return schedule


def polynomial_schedule(
    a: float = 1.0, b: float = 1.0, gamma: float = 0.55
) -> StepSchedule:
    """Create a polynomial step schedule.

    Parameters
    ----------
    a: float
        Scale of all step sizes.
    b: float
        The stabilization constant.
    gamma: float
        The decay rate :math:`\gamma \in (0.5, 1.0]`.

    Returns
    -------
    schedule_fn: StepSchedule
    """

    if not 0.5 < gamma <= 1.0:
        raise ValueError("`gamma` should be in (0.5, 1.0] range.")

    def schedule(step: Array):
        return a * (b + step) ** (-gamma)

    return schedule


def constant_schedule_with_cosine_burnin(
    init_step_size: float, final_step_size: float, burnin_steps: int
) -> StepSchedule:
    """Create a constant schedule with cosine burn-in.

    Parameters
    ----------
    init_step_size: float
        The initial step size.
    final_step_size: float
        The desired final step size.
    burnin_steps: int
        The length of burn-in, in steps.

    Returns
    -------
    schedule_fn: StepSchedule
    """
    if not init_step_size >= 0:
        raise ValueError("`init_step_size` should be >= 0.")
    if not final_step_size >= 0:
        raise ValueError("`final_step_size` should be >= 0.")
    if not burnin_steps >= 0:
        raise ValueError("`burnin_steps` should be >= 0.")

    def schedule(step: Array):
        t = jnp.minimum(step / burnin_steps, 1.0)
        coef = (1 + jnp.cos(t * np.pi)) * 0.5
        return coef * init_step_size + (1 - coef) * final_step_size

    return schedule


def cyclical_cosine_schedule_with_const_burnin(
    init_step_size: float, burnin_steps: int, cycle_length: int
) -> StepSchedule:
    """Create a cyclical cosine schedule with constant burn-in.

    Parameters
    ----------
    init_step_size: float
        The initial step size.
    burnin_steps: int
        The length of burn-in, in steps.
    cycle_length: int
        The length of the cosine cycle, in steps.

    Returns
    -------
    schedule_fn: StepSchedule
    """
    if not init_step_size >= 0:
        raise ValueError("`init_step_size` should be >= 0.")
    if not burnin_steps >= 0:
        raise ValueError("`burnin_steps` should be >= 0.")
    if not cycle_length >= 0:
        raise ValueError("`cycle_length` should be >= 0.")

    def schedule(step: Array):
        t = jnp.maximum(step - burnin_steps - 1, 0.0)
        t = (t % cycle_length) / cycle_length
        return 0.5 * init_step_size * (1 + jnp.cos(t * np.pi))

    return schedule



from __future__ import annotations

from typing import (
    Optional,
    Tuple,
)

import jax.numpy as jnp

from fortuna.conformal.multivalid.iterative.base import IterativeMultivalidMethod
from fortuna.conformal.multivalid.mixins.multicalibrator import MulticalibratorMixin
from fortuna.typing import Array


class Multicalibrator(MulticalibratorMixin, IterativeMultivalidMethod):
    def __init__(self, seed: int = 0):
        """
        A multicalibration method that provides multivalid coverage guarantees. See Algorithm 15 in `Aaron Roth's notes
        <https://www.cis.upenn.edu/~aaroth/uncertainty-notes.pdf>`_.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)
        self._patch_list = []

    def _calibration_error(
        self,
        v: Array,
        g: Array,
        c: Array,
        tau: Array,
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        **kwargs,
    ):
        patch, b, prob_b = self.__calibration_error(
            v=v,
            g=g,
            c=c,
            b=None,
            tau=tau,
            scores=scores,
            groups=groups,
            values=values,
            buckets=buckets,
        )
        return prob_b * patch**2, b

    def __calibration_error(
        self,
        v: Array,
        g: Array,
        c: Array,
        b: Optional[Array],
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
    ) -> Tuple[Array, Array, Array]:
        ey, b, prob_b = self._compute_expectation(
            v=v,
            g=g,
            c=c,
            b=b,
            tau=tau,
            scores=scores,
            groups=groups,
            values=values,
            n_buckets=len(buckets),
        )
        patch = jnp.where(prob_b > 0, ey - self._get_mean_values(values, b, c), 0.0)
        return patch, b, prob_b

    def _compute_expectation(
        self,
        v: Array,
        g: Array,
        c: Array,
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        n_buckets: int,
        b: Optional[Array] = None,
    ):
        if b is None:
            b = self._get_b(
                groups=groups,
                values=values,
                v=v,
                g=g,
                c=c,
                tau=tau,
                n_buckets=n_buckets,
            )
        filtered_scores = scores * b
        prob_b = jnp.mean(b)
        mean = jnp.mean(filtered_scores) / prob_b
        return mean, b, prob_b

    def _get_patch(
        self,
        v: Array,
        g: Array,
        c: Array,
        b: Optional[Array],
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        patch_type: str,
        **kwargs,
    ) -> Array:
        if patch_type == "additive":
            return self.__calibration_error(
                v=v,
                g=g,
                c=c,
                b=b,
                tau=tau,
                scores=scores,
                groups=groups,
                values=values,
                buckets=buckets,
            )[0]
        if patch_type == "multiplicative":
            ey, b, prob_b = self._compute_expectation(
                v=v,
                g=g,
                c=c,
                b=b,
                tau=tau,
                scores=scores,
                groups=groups,
                values=values,
                n_buckets=len(buckets),
            )
            return ey / self._get_mean_values(values, b, c)

    @staticmethod
    def _get_mean_values(values: Array, b: Array, c: Array):
        if values.ndim == 1:
            return jnp.mean(values * b) / jnp.mean(b)
        return jnp.mean(values[:, c] * b) / jnp.mean(b)



import jax.numpy as jnp

from fortuna.conformal.multivalid.mixins.multicalibrator import MulticalibratorMixin
from fortuna.conformal.multivalid.one_shot.base import OneShotMultivalidMethod
from fortuna.typing import Array


class OneShotMulticalibrator(MulticalibratorMixin, OneShotMultivalidMethod):
    def __init__(self, seed: int = 0):
        super().__init__(seed=seed)

    def _get_patch(
        self, v: Array, c: Array, scores: Array, values: Array, min_prob_b: float
    ) -> Array:
        return self._compute_expectation(
            v=v, c=c, scores=scores, values=values, min_prob_b=min_prob_b
        )

    def _compute_expectation(
        self, v: Array, c: Array, scores: Array, values: Array, min_prob_b: float
    ):
        b = self._get_b(values=values, v=v, c=c, n_buckets=self.n_buckets)
        filtered_scores = scores * b
        prob_b = jnp.mean(b)
        mean = jnp.where(prob_b > min_prob_b, jnp.mean(filtered_scores) / prob_b, v)
        return mean



from typing import (
    Any,
    Dict,
    Optional,
)

from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.sngp import SNGP_NAME


class SNGPPosteriorApproximator(PosteriorApproximator):
    def __init__(
        self,
        *args,
        output_dim: int,
        gp_hidden_features: int = 1024,
        normalize_input: bool = False,
        ridge_penalty: float = 1.0,
        momentum: Optional[float] = None,
        mean_field_factor: float = 1.0,
        **kwargs,
    ):
        """
        SNGP posterior approximator. It is responsible to define how the posterior distribution is approximated.

        Parameters
        ----------
        output_dim: int
            The output dimension of the network.
        normalize_input: bool
            Whether to normalize the input using nn.LayerNorm.
        gp_hidden_features: int
            The number of random fourier features.
        ridge_penalty: float
            Initial Ridge penalty to weight covariance matrix.
            This value is used to stabilize the eigenvalues of weight covariance estimate :math:`\Sigma` so that
            the matrix inverse can be computed for :math:`\Sigma = (\mathbf{I}*s+\mathbf{X}^T\mathbf{X})^{-1}`.
            The ridge factor :math:`s` cannot be too large since otherwise it will dominate
            making the covariance estimate not meaningful.
        momentum: Optional[float]
            A discount factor used to compute the moving average for posterior
            precision matrix. Analogous to the momentum factor in batch normalization.
            If `None` then update covariance matrix using a naive sum without
            momentum, which is desirable if the goal is to compute the exact
            covariance matrix by passing through data once (say in the final epoch).
            In this case, make sure to reset the precision matrix variable between
            epochs to avoid double counting.
        mean_field_factor: float
            The scale factor for mean-field approximation, used to adjust (at inference time) the influence of
            posterior variance in posterior mean approximation.
            See `Zhiyun L. et al., 2020 <https://arxiv.org/abs/2006.07584>`_ for more details.
        mean_field_factor: float
            The scale factor for mean-field approximation, used to adjust (at inference time) the influence of
            posterior variance in posterior mean approximation.
            See `Zhiyun L. et al., 2020 <https://arxiv.org/abs/2006.07584>`_ for more details.
        """
        super(SNGPPosteriorApproximator, self).__init__(*args, **kwargs)
        self.output_dim = output_dim
        self.gp_hidden_features = gp_hidden_features
        self.normalize_input = normalize_input
        self.ridge_penalty = ridge_penalty
        self.momentum = momentum
        self.mean_field_factor = mean_field_factor

    def __str__(self):
        return SNGP_NAME

    @property
    def posterior_method_kwargs(self) -> Dict[str, Any]:
        return {
            "output_dim": self.output_dim,
            "gp_hidden_features": self.gp_hidden_features,
            "normalize_input": self.normalize_input,
            "ridge_penalty": self.ridge_penalty,
            "momentum": self.momentum,
            "mean_field_factor": self.mean_field_factor,
        }



import abc
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

from jax import (
    jit,
    pmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.data.loader import (
    DataLoader,
    DeviceDimensionAugmentedLoader,
    InputsLoader,
)
from fortuna.model.model_manager.base import ModelManager
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.base import ProbOutputLayer
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
)
from fortuna.utils.random import WithRNG


class Likelihood(WithRNG):
    def __init__(
        self,
        model_manager: ModelManager,
        prob_output_layer: ProbOutputLayer,
        output_calib_manager: OutputCalibManager,
    ):
        """
        A likelihood function abstract class. In this class, the likelihood function is additionally assumed to be a
        probability density function, i.e. positive and integrating to 1. The likelihood is formed by three objects
        applied in sequence: the model manager, the output calibrator and the probabilistic output layer. The output
        maker maps parameters and inputs to outputs. The output calibration takes outputs and returns some calibrated
        version of them. The probabilistic output layer describes the probability distribution of the calibrated
        outputs.

        Parameters
        ----------
        model_manager : ModelManager
            An model manager. This objects orchestrates the evaluation of the models.
        prob_output_layer : ProbOutputLayer
            A probabilistic output layer object. This object characterizes the probability distribution of the
            target variable given the calibrated outputs.
        output_calib_manager : OutputCalibManager
            An output calibration manager object. It transforms outputs of the model manager into some
            calibrated version of them.
        """
        self.model_manager = model_manager
        self.prob_output_layer = prob_output_layer
        self.output_calib_manager = output_calib_manager

    def log_prob(
        self,
        params: Params,
        data_loader: DataLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Evaluate the log-likelihood function.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        data_loader : DataLoader
            A data loader.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            The evaluation of the log-likelihood function.
        """
        return self._loop_fun_through_data_loader(
            self._batched_log_prob,
            params,
            data_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    def _batched_log_prob(
        self,
        params: Params,
        batch: Batch,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = self._get_batched_calibrated_outputs(
            params, batch[0], mutable, calib_params, calib_mutable, **kwargs
        )
        return self.prob_output_layer.log_prob(outputs, batch[1], **kwargs)

    def _batched_log_joint_prob(
        self,
        params: Params,
        batch: Batch,
        n_data: int,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        train: bool = False,
        outputs: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Any]]:
        """
        Evaluate the batched log-likelihood function.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        batch : Batch
            A batch of data points.
        n_data : int
            The total number of data points over which the likelihood is joint. This is used to rescale the batched
            log-likelihood function to better approximate the full likelihood.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        return_aux : Optional[List[str]]
            The auxiliary objects to return. We support 'outputs', 'mutable' and 'calib_mutable'. If this argument is
            not given, no auxiliary object is returned.
        train : bool
            Whether the method is called during training.
        outputs : Optional[jnp.ndarray]
            Pre-computed batch of outputs.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, Any]]
            The evaluation of the batched log-likelihood function. If `return_aux` is given, the corresponding
            auxiliary objects are also returned.
        """
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "mutable", "calib_mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise AttributeError(
                """The auxiliary objects {} is unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )
        if train and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `train` must be set to `False`."""
            )
        if "mutable" in return_aux and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `return_aux` cannot contain 'mutable'`."""
            )
        if not train and "mutable" in return_aux:
            raise ValueError(
                "Returning an auxiliary mutable is supported only during training. Please either set `train` to "
                "`True`, or remove 'mutable' from `return_aux`."
            )
        if "mutable" in return_aux and mutable is None:
            raise ValueError(
                "In order to be able to return an auxiliary mutable, an initial mutable must be passed as `mutable`. "
                "Please either remove 'mutable' from `return_aux`, or pass an initial mutable as `mutable`."
            )
        if "mutable" not in return_aux and mutable is not None and train is True:
            raise ValueError(
                """You need to add `mutable` to `return_aux`. When you provide a (not null) `mutable` variable during
                training, that variable will be updated during the forward pass."""
            )

        inputs, targets = batch
        if outputs is None:
            outs = self.model_manager.apply(
                params,
                inputs,
                train=train,
                mutable=mutable,
                rng=rng,
            )
            if "mutable" in return_aux:
                outputs, aux = outs
                mutable = aux["mutable"]
            else:
                outputs = outs

        aux = dict()
        if self.output_calib_manager is not None:
            outs = self.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
                calib="calib_mutable" in return_aux,
            )
        if (
            calib_mutable is not None
            and calib_mutable["output_calibrator"] is not None
            and "calib_mutable" in return_aux
        ):
            outputs, aux["calib_mutable"] = outs
            aux["calib_mutable"] = dict(output_calibrator=aux["calib_mutable"])
        else:
            outputs = outs
            if "calib_mutable" in return_aux:
                aux["calib_mutable"] = dict(output_calibrator=None)

        log_joint_prob = jnp.sum(
            self.prob_output_layer.log_prob(outputs, targets, train=train, **kwargs)
        )
        batch_weight = n_data / targets.shape[0]
        log_joint_prob *= batch_weight

        if len(return_aux) == 0:
            return log_joint_prob
        else:
            if "outputs" in return_aux:
                aux["outputs"] = outputs
            if "mutable" in return_aux:
                aux["mutable"] = mutable
            return log_joint_prob, aux

    def sample(
        self,
        n_target_samples: int,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
        """
        Sample target variables from the likelihood function for each input variable.

        Parameters
        ----------
        n_target_samples : int
            The number of samples to draw from the likelihood for each input data point.
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        return_aux : Optional[List[str]]
            The auxiliary objects to return. We support 'outputs'. If this argument is not given, no auxiliary object
            is returned.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]
            The samples of the target variable for each input. If `return_aux` is given, the corresponding auxiliary
            objects are also returned.
        """
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise Exception(
                """The auxiliary objects {} are unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )

        outputs = self.get_calibrated_outputs(
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

        samples = self.prob_output_layer.sample(
            n_target_samples, outputs, rng=rng, **kwargs
        )
        if len(return_aux) > 0:
            return samples, dict(outputs=outputs)
        return samples

    def _batched_sample(
        self,
        n_target_samples: int,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise Exception(
                """The auxiliary objects {} are unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )

        outputs = self._get_batched_calibrated_outputs(
            params, inputs, mutable, calib_params, calib_mutable, **kwargs
        )

        samples = self.prob_output_layer.sample(
            n_target_samples, outputs, rng=rng, **kwargs
        )
        if len(return_aux) > 0:
            return samples, dict(outputs=outputs)
        return samples

    def get_calibrated_outputs(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Compute the outputs and their calibrated version.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            The calibrated outputs.
        """
        outputs = self.get_outputs(params, inputs_loader, mutable, distribute, **kwargs)

        if (
            self.output_calib_manager is not None
            and self.output_calib_manager.output_calibrator is not None
        ):
            outputs = self.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
                **kwargs,
            )
        return outputs

    def _get_batched_calibrated_outputs(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = self.model_manager.apply(params, inputs, mutable, **kwargs)

        if (
            self.output_calib_manager is not None
            and self.output_calib_manager.output_calibrator is not None
        ):
            outputs = self.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
                **kwargs,
            )
        return outputs

    def get_outputs(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Compute the outputs and their calibrated version.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            The calibrated outputs.
        """

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)

        @jit
        def fun(_inputs):
            return self.model_manager.apply(params, _inputs, mutable, **kwargs)

        outputs = []
        for inputs in inputs_loader:
            outputs.append(
                self._unshard_array(pmap(fun)(inputs)) if distribute else fun(inputs)
            )
        return jnp.concatenate(outputs, 0)

    def mean(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood mean of the target variable, that is

        .. math::
            \mathbb{E}_{Y|w, x}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood mean for each input.
        """
        return self._loop_fun_through_inputs_loader(
            self._batched_mean,
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    @abc.abstractmethod
    def _batched_mean(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def mode(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood mode of the target variable, that is

        .. math::
            \text{argmax}_y\ p(y|w, x),

        where:
         - :math:`x` is an observed input variable;
         - :math:`w` denotes the observed model parameters;
         - :math:`y` is the target variable to optimize upon.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood mode for each input.
        """
        return self._loop_fun_through_inputs_loader(
            self._batched_mode,
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    @abc.abstractmethod
    def _batched_mode(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def variance(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood variance of the target variable, that is

        .. math::
            \text{Var}_{Y|w,x}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood variance for each input.
        """
        return self._loop_fun_through_inputs_loader(
            self._batched_variance,
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    @abc.abstractmethod
    def _batched_variance(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def std(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        variance: Optional[jnp.ndarray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood standard deviation of the target variable, that is

        .. math::
            \sqrt{\text{Var}_{Y|w,x}[Y]},

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        variance : Optional[jnp.ndarray]
            An estimate of the likelihood variance for each input.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood standard deviation for each input.
        """
        if variance is None:
            variance = self.variance(
                params,
                inputs_loader,
                mutable,
                calib_params=calib_params,
                calib_mutable=calib_mutable,
                distribute=distribute,
                **kwargs,
            )
        return jnp.sqrt(variance)

    @abc.abstractmethod
    def entropy(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood entropy, that is

        .. math::
            -\mathbb{E}_{Y|w,x}[\log p(Y|w,x)]

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood entropy for each input.
        """
        pass

    @staticmethod
    def _unshard_array(arr: Array) -> Array:
        return arr.reshape((arr.shape[0] * arr.shape[1],) + arr.shape[2:])

    def _loop_fun_through_inputs_loader(
        self,
        fun: Callable,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> Array:
        def fun2(_inputs):
            return fun(params, _inputs, mutable, calib_params, calib_mutable, **kwargs)

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
            fun2 = pmap(fun2)
            return jnp.concatenate(
                [self._unshard_array(fun2(inputs)) for inputs in inputs_loader], 0
            )
        fun2 = jit(fun2)
        return jnp.concatenate([fun2(inputs) for inputs in inputs_loader], 0)

    def _loop_fun_through_data_loader(
        self,
        fun: Callable,
        params: Params,
        data_loader: DataLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> Array:
        def fun2(_batch):
            return fun(params, _batch, mutable, calib_params, calib_mutable, **kwargs)

        if distribute:
            data_loader = DeviceDimensionAugmentedLoader(data_loader)
            fun2 = pmap(fun2)
            return jnp.concatenate(
                [self._unshard_array(fun2(batch)) for batch in data_loader], 0
            )
        fun2 = jit(fun2)
        return jnp.concatenate([fun2(batch) for batch in data_loader], 0)



from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

import jax.numpy as jnp
from jax.tree_util import tree_map
import numpy as np

from fortuna.typing import Array


def convert_string_to_tuple(s: str) -> Tuple:
    return tuple([ord(c) for c in s])


def convert_string_to_np_array(s: str) -> np.ndarray:
    return np.array([ord(c) for c in s])


def encode_tuple_of_lists_of_strings_to_numpy(
    a: Optional[Tuple[List[str]]],
) -> Optional[Tuple[List[Array]]]:
    return (
        tuple([[convert_string_to_np_array(s) for s in key_path] for key_path in a])
        if a is not None
        else None
    )


def decode_encoded_tuple_of_lists_of_strings_to_array(
    encoded: Optional[Dict[str, List[Array]]]
) -> Optional[Tuple[List[str], ...]]:
    if encoded is None:
        return None
    encoded = tree_map(lambda v: "".join([chr(o) for o in v]), encoded)
    if isinstance(encoded, dict):
        return tuple([list(v.values()) for k, v in encoded.items()])
    else:
        return encoded



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
)

from flax.core import FrozenDict

from fortuna.training.train_state import TrainState
from fortuna.typing import (
    CalibMutable,
    CalibParams,
    Mutable,
    OptaxOptimizer,
    Params,
)


class CalibState(TrainState):
    params: Params
    mutable: Optional[Mutable] = None
    calib_params: Optional[CalibParams] = None
    calib_mutable: Optional[CalibMutable] = None

    @classmethod
    def init(
        cls,
        params: Params,
        mutable: Optional[Mutable] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> Any:
        return cls(
            apply_fn=None,
            params=params,
            opt_state=kwargs["opt_state"]
            if optimizer is None and "opt_state" in kwargs
            else optimizer.init(params),
            mutable=mutable,
            step=kwargs.get("step", 0),
            tx=optimizer,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            **{
                k: v
                for k, v in kwargs.items()
                if k not in ["opt_state", "apply_fn", "tx", "step"] and hasattr(cls, k)
            },
        )

    @classmethod
    def init_from_dict(
        cls,
        d: Dict,
        optimizer: Optional[OptaxOptimizer] = None,
        **kwargs,
    ) -> CalibState:
        kwargs = {
            **kwargs,
            **{
                k: v
                for k, v in d.items()
                if k
                not in [
                    "params",
                    "mutable",
                    "optimizer",
                    "calib_params",
                    "calib_mutable",
                ]
            },
        }
        return cls.init(
            FrozenDict(d["params"]),
            FrozenDict(d["mutable"]) if d["mutable"] is not None else None,
            optimizer,
            FrozenDict(d.get("calib_params"))
            if d["calib_params"] is not None
            else None,
            FrozenDict(d.get("calib_mutable"))
            if d["calib_mutable"] is not None
            else None,
            **kwargs,
        )



from typing import List

import jax.numpy as jnp
import numpy as np

from fortuna.conformal.classification.base import ConformalClassifier
from fortuna.conformal.multivalid.iterative.batch_mvp import BatchMVPConformalMethod
from fortuna.typing import Array


class BatchMVPConformalClassifier(BatchMVPConformalMethod, ConformalClassifier):
    def __init__(self, seed: int = 0):
        """
        This class implements a classification version of BatchMVP
        `[Jung et al., 2022] <https://arxiv.org/abs/2209.15145>`_,
        a multivalid conformal prediction method that satisfies coverage guarantees conditioned on group membership
        and non-conformity threshold.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)

    @staticmethod
    def conformal_set(
        class_scores: Array,
        thresholds: Array,
    ) -> List[List[int]]:
        """
        Compute a conformal set for each input.

        Parameters
        ----------
        class_scores: Array
            A two-dimensional array of scores. The first dimension is over the different inputs.
            The second dimension is over all the possible classes. For example, if there are 10 classes,
            the first row of `class_scores` show be :math:`[s(x_1, 0), \dots, s(x_1, 9)]`.
        thresholds: Array
            A one-dimensional array of thresholds over the different inputs. This should be obtained from the `calibrate`
            method.

        Returns
        -------
        List[List[int]]
            Conformal sets for each input data point.
        """
        if class_scores.ndim != 2:
            raise ValueError(
                "`class_scores` must bse a 2-dimensional array. "
                "The first dimension is over the different inputs. "
                "The second dimension is over all the possible classes."
            )
        if thresholds.ndim != 1:
            raise ValueError("`thresholds` must be a 1-dimensional array.")
        if class_scores.shape[0] != thresholds.shape[0]:
            raise ValueError(
                "The first dimension of `class_scores` and `thresholds` must be over the same input data "
                "points."
            )
        bools = class_scores <= thresholds[:, None]

        sizes = np.sum(bools, 1)
        sets = np.zeros(bools.shape[0], dtype=object)
        for s in np.unique(sizes):
            idx = jnp.where(sizes == s)[0]
            sets[idx] = np.nonzero(bools[idx])[1].reshape(len(idx), s).tolist()
        return sets.tolist()



import pathlib
from typing import (
    Optional,
    Tuple,
    Type,
)

from jax import (
    pure_callback,
    random,
)
from jax._src.prng import PRNGKeyArray

from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_posterior_state_repository import (
    SGMCMCPosteriorStateRepository,
)
from fortuna.typing import Path
from fortuna.utils.strings import decode_encoded_tuple_of_lists_of_strings_to_array


class SGMCMCPosterior(Posterior):
    """Base SGMCMC posterior approximators class."""

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = pure_callback(
            lambda j: self.state.get(i=j),
            self.state.get(i=0),
            random.choice(rng, self.posterior_approximator.n_samples),
        )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def load_state(self, checkpoint_dir: Path) -> None:
        try:
            state = self.restore_checkpoint(pathlib.Path(checkpoint_dir) / "c")
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_dir={checkpoint_dir}`."
            )
        which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
            state._encoded_which_params
        )
        self.state = SGMCMCPosteriorStateRepository(
            size=self.posterior_approximator.n_samples,
            checkpoint_dir=checkpoint_dir,
            which_params=which_params,
            all_params=state.params if which_params else None,
        )

    def save_state(self, checkpoint_dir: Path, keep_top_n_checkpoints: int = 1) -> None:
        for i in range(self.posterior_approximator.n_samples):
            self.state.put(state=self.state.get(i), i=i, keep=keep_top_n_checkpoints)

    def _restore_state_from_somewhere(
        self,
        fit_config: FitConfig,
        allowed_states: Optional[Tuple[Type[MAPState], ...]] = None,
    ) -> MAPState:
        if fit_config.checkpointer.restore_checkpoint_path is not None:
            restore_checkpoint_path = (
                pathlib.Path(fit_config.checkpointer.restore_checkpoint_path) / "c"
            )
            state = self.restore_checkpoint(
                restore_checkpoint_path=restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )
        elif fit_config.checkpointer.start_from_current_state is not None:
            state = self.state.get(
                i=self.state.size - 1,
                optimizer=fit_config.optimizer.method,
            )

        if allowed_states is not None and not isinstance(state, allowed_states):
            raise ValueError(
                f"The type of the restored checkpoint must be within {allowed_states}. "
                f"However, {fit_config.checkpointer.restore_checkpoint_path} pointed to a state "
                f"with type {type(state)}."
            )

        self._check_state(state)
        return state



from __future__ import annotations

import logging
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from flax.training.common_utils import (
    shard,
    shard_prng_key,
)
import jax
from jax import (
    devices,
    hessian,
    jit,
    lax,
    pmap,
    vjp,
)
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
from jax.tree_util import tree_map
import tqdm

from fortuna.data.loader import (
    DataLoader,
    DeviceDimensionAugmentedLoader,
)
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME
from fortuna.prob_model.posterior.laplace.laplace_approximator import (
    LaplacePosteriorApproximator,
)
from fortuna.prob_model.posterior.laplace.laplace_state import LaplaceState
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.prob_model.posterior.run_preliminary_map import run_preliminary_map
from fortuna.prob_model.prior.gaussian import (
    DiagonalGaussianPrior,
    IsotropicGaussianPrior,
)
from fortuna.typing import (
    AnyKey,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
    Status,
)
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
    nested_unpair,
)
from fortuna.utils.random import generate_random_normal_like_tree
from fortuna.utils.strings import decode_encoded_tuple_of_lists_of_strings_to_array


class LaplacePosterior(Posterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: LaplacePosteriorApproximator,
    ):
        """
        Laplace approximation posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: LaplacePosteriorApproximator
            A Laplace posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)
        if type(joint.prior) not in [DiagonalGaussianPrior, IsotropicGaussianPrior]:
            raise ValueError(
                """The Laplace posterior_approximation is not supported for this model. The prior distribution must be one of the
                following choices: {}.""".format(
                    [DiagonalGaussianPrior, IsotropicGaussianPrior]
                )
            )

    def __str__(self):
        return LAPLACE_NAME

    def _gnn_approx(
        self,
        params: Params,
        train_data_loader: DataLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        which_params: Optional[Tuple[List[AnyKey, ...]]] = None,
        factorization: str = "diagonal",
        verbose: bool = True,
    ) -> Params:
        """
        Estimate a standard deviation for each parameter using a diagonal Generalized Gauss-Newton Hessian
        approximation.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        train_data_loader: DataLoader
            A training data loader.
        mutable: Optional[Mutable]
            Mutable objects.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable] = None
            The calibration mutable objects used to evaluate the calibrators.
        which_params : Optional[Tuple[List[AnyKey, ...]]]
            Sequences of keys indicating which parameters to compute the Hessian upon.
        factorization: str = "diagonal"
            Factorization of the GGN approximation. Currently, only "diagonal" is supported.
        verbose: bool
            Whether to log the training progress.

        Returns
        -------
        Params
            An estimate of the likelihood standard deviation for each random parameter.
        """
        rav, unravel = ravel_pytree(
            tuple([nested_get(params, keys) for keys in which_params])
            if which_params
            else params
        )

        def get_params_from_rav(_rav):
            unrav = unravel(_rav)
            return (
                FrozenDict(
                    nested_set(
                        params.unfreeze(),
                        which_params,
                        unrav,
                    )
                )
                if which_params
                else unrav
            )

        def apply_calib_model_manager(_params, _batch_inputs):
            outputs = self.joint.likelihood.model_manager.apply(
                _params, _batch_inputs, mutable=mutable, train=False
            )
            outputs = self.joint.likelihood.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
            )
            return outputs

        apply_fn = lambda _rav, x: apply_calib_model_manager(
            get_params_from_rav(_rav), x
        ).squeeze(0)
        vjp_fn = lambda x: vjp(lambda _rav: apply_fn(_rav, x), rav)[1]

        def eig_hess_fn(vars):
            hess = hessian(
                lambda __o: self.joint.likelihood.prob_output_layer.log_prob(
                    __o, vars[1]
                )
            )(vars[0])
            return jnp.linalg.eigh(hess)

        def compute_hess_batch(_batch_inputs, _batch_targets):
            lam, z = lax.map(
                eig_hess_fn,
                (apply_calib_model_manager(params, _batch_inputs), _batch_targets),
            )
            ztj = lax.map(
                lambda v: lax.map(vjp_fn(v[0]), v[1].T),
                (tree_map(lambda x: x[:, None], _batch_inputs), z),
            )[0]
            if factorization == "diagonal":
                return -jnp.sum(lam[:, :, None] * ztj**2, (0, 1))
            raise ValueError(
                f"`factorization={factorization}` not recognized. Currently, only "
                f"`factorization='diagonal'` is supported."
            )

        n_gpu_devices = len([d for d in devices() if d.platform == "gpu"])
        if n_gpu_devices > 0:
            train_data_loader = DeviceDimensionAugmentedLoader(train_data_loader)
            compute_hess_batch = pmap(compute_hess_batch, axis_name="batch")
        else:
            compute_hess_batch = jit(compute_hess_batch)

        h = 0.0
        for i, (batch_inputs, batch_targets) in enumerate(train_data_loader):
            if verbose:
                logging.info(f"Hessian approximation for batch {i + 1}.")
            h += compute_hess_batch(batch_inputs, batch_targets)

        if n_gpu_devices > 0:
            h = jnp.sum(h, 0)
        return unravel(h)

    def _compute_std(self, prior_log_var: float, hess_lik_diag: Params) -> Params:
        hess_prior = jnp.exp(-prior_log_var)
        hess_lik_diag_rav, unravel = ravel_pytree(hess_lik_diag)
        return unravel(1 / jnp.sqrt(hess_prior + hess_lik_diag_rav))

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> Dict[str, Status]:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        status = dict()

        if super()._is_state_available_somewhere(fit_config):
            state = super()._restore_state_from_somewhere(
                fit_config=fit_config, allowed_states=(MAPState, LaplaceState)
            )

        elif super()._should_run_preliminary_map(fit_config, map_fit_config):
            state, status["map"] = run_preliminary_map(
                joint=self.joint,
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                map_fit_config=map_fit_config,
                rng=self.rng,
                **kwargs,
            )
        else:
            raise ValueError(
                "The Laplace approximation must start from a preliminary run of MAP or an existing "
                "checkpoint or state. Please configure `map_fit_config`, or "
                "`fit_config.checkpointer.restore_checkpoint_path`, "
                "or `fit_config.checkpointer.start_from_current_state`."
            )

        state = self._init_map_state(state=state, fit_config=fit_config)

        if fit_config.optimizer.freeze_fun is not None:
            which_params = get_trainable_paths(
                params=state.params, freeze_fun=fit_config.optimizer.freeze_fun
            )
        else:
            which_params = None

        logging.info("Run the Laplace approximation.")
        hess_lik_diag = self._gnn_approx(
            state.params,
            train_data_loader,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
            which_params=which_params,
            verbose=fit_config.monitor.verbose,
        )
        state = LaplaceState.convert_from_map_state(
            map_state=state,
            hess_lik_diag=hess_lik_diag,
            prior_log_var=self.joint.prior.log_var,
            which_params=which_params,
        )

        if fit_config.checkpointer.save_checkpoint_dir:
            self.save_checkpoint(
                state,
                save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
                keep=fit_config.checkpointer.keep_top_n_checkpoints,
                force_save=True,
            )

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        if (
            val_data_loader is not None
            and self.posterior_approximator.tune_prior_log_variance
        ):
            logging.info("Tuning the prior log-variance now")
            opt_prior_log_var = self.prior_log_variance_tuning(
                val_data_loader=val_data_loader,
                n_posterior_samples=5,
                distribute=fit_config.processor.devices == -1,
            )
            state = state.replace(prior_log_var=opt_prior_log_var)
            self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
            logging.info(f"Best prior log-variance found: {opt_prior_log_var}")
        return status

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> JointState:
        if rng is None:
            rng = self.rng.get()
        state: LaplaceState = self.state.get()
        if kwargs.get("prior_log_var") is not None:
            state = state.replace(prior_log_var=kwargs.get("prior_log_var"))

        if state._encoded_which_params is not None:
            which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
                state._encoded_which_params
            )
            mean, hess_lik_diag = nested_unpair(
                state.params.unfreeze(),
                which_params,
                ("mean", "hess_lik_diag"),
            )
            std = self._compute_std(
                prior_log_var=state.prior_log_var, hess_lik_diag=hess_lik_diag
            )

            noise = generate_random_normal_like_tree(rng, std)
            params = nested_set(
                d=mean,
                key_paths=which_params,
                objs=tuple(
                    [
                        tree_map(
                            lambda m, s, e: m + s * e,
                            nested_get(mean, keys),
                            nested_get(std, keys),
                            nested_get(noise, keys),
                        )
                        for keys in which_params
                    ]
                ),
            )
            for k, v in params.items():
                params[k] = FrozenDict(v)
            state = state.replace(params=FrozenDict(params))
        else:
            mean, hess_lik_diag = dict(), dict()
            for k, v in state.params.items():
                mean[k] = FrozenDict({"params": v["params"]["mean"]})
                hess_lik_diag[k] = FrozenDict({"params": v["params"]["hess_lik_diag"]})

            std = self._compute_std(
                prior_log_var=state.prior_log_var, hess_lik_diag=hess_lik_diag
            )
            state = state.replace(
                params=FrozenDict(
                    tree_map(
                        lambda m, s, e: m + s * e,
                        mean,
                        std,
                        generate_random_normal_like_tree(rng, std),
                    )
                )
            )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def _init_map_state(
        self, state: Union[MAPState, LaplaceState], fit_config: FitConfig
    ) -> MAPState:
        if isinstance(state, LaplaceState):
            if state._encoded_which_params is not None:
                which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
                    state._encoded_which_params
                )
                state = state.replace(
                    params=FrozenDict(
                        nested_unpair(
                            d=state.params.unfreeze(),
                            key_paths=which_params,
                            labels=("mean", "hess_lik_diag"),
                        )[0]
                    )
                )
            else:
                state = state.replace(
                    params=FrozenDict(
                        {
                            k: dict(params=v["params"]["mean"])
                            for k, v in state.params.items()
                        }
                    )
                )

        state = MAPState.init(
            params=state.params,
            mutable=state.mutable,
            optimizer=fit_config.optimizer.method,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

        return state

    def _batched_log_prob(
        self,
        batch,
        prior_log_var: float,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
        import jax.random as random
        import jax.scipy as jsp

        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def _lik_log_batched_prob(key):
            sample = self.sample(inputs=batch[0], rng=key, prior_log_var=prior_log_var)
            return self.joint.likelihood._batched_log_prob(
                sample.params,
                batch,
                mutable=sample.mutable,
                calib_params=sample.calib_params,
                calib_mutable=sample.calib_mutable,
                **kwargs,
            )

        return jsp.special.logsumexp(
            lax.map(_lik_log_batched_prob, keys), axis=0
        ) - jnp.log(n_posterior_samples)

    def prior_log_variance_tuning(
        self,
        val_data_loader: DataLoader,
        n_posterior_samples: int = 10,
        mode: str = "cv",
        min_prior_log_var: float = -3,
        max_prior_log_var: float = 3,
        grid_size: int = 20,
        distribute: bool = False,
    ) -> jnp.ndarray:
        if mode == "cv":
            return self._prior_log_variance_tuning_cv(
                val_data_loader,
                n_posterior_samples,
                min_prior_log_var,
                max_prior_log_var,
                grid_size,
                distribute,
            )
        elif mode == "marginal_lik":
            raise NotImplementedError(
                f"Optimizing the prior log variance via marginal likelihood maximization is not yet available."
            )
        else:
            raise ValueError(f"Unrecognized mode={mode} for prior log variance tuning.")

    def _prior_log_variance_tuning_cv(
        self,
        val_data_loader: DataLoader,
        n_posterior_samples: int,
        min_prior_log_var: float,
        max_prior_log_var: float,
        grid_size: int,
        distribute: bool,
    ) -> jnp.ndarray:
        best = None
        candidates = list(
            jnp.linspace(min_prior_log_var, max_prior_log_var, grid_size)
        ) + [jnp.array(self.joint.prior.log_var)]
        if distribute:
            rng = shard_prng_key(jax.random.PRNGKey(0))
            val_data_loader = DeviceDimensionAugmentedLoader(val_data_loader)
            candidates = [shard(c) for c in candidates]
            fn = pmap(self._batched_log_prob, static_broadcasted_argnums=(2,))
        else:
            fn = jit(self._batched_log_prob, static_argnums=(2,))

        for lpv in tqdm.tqdm(candidates, desc="Tuning prior log-var"):
            neg_log_prob = -jnp.sum(
                jnp.concatenate(
                    [
                        self.joint.likelihood._unshard_array(
                            fn(batch, lpv, n_posterior_samples, rng)
                        )
                        for batch in val_data_loader
                    ],
                    0,
                )
            )
            if best is None or neg_log_prob < best[-1]:
                best = (lpv, neg_log_prob)

        opt_prior_log_var = best[0].reshape()
        return opt_prior_log_var



from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

from jax import vmap
import jax.numpy as jnp

from fortuna.conformal.classification.base import ConformalClassifier
from fortuna.conformal.multivalid.iterative.base import IterativeMultivalidMethod
from fortuna.conformal.multivalid.mixins.batchmvp import BatchMVPMixin
from fortuna.typing import Array


class BatchMVPConformalMethod(
    BatchMVPMixin, IterativeMultivalidMethod, ConformalClassifier
):
    def __init__(self, seed: int = 0):
        """
        This class implements a classification version of BatchMVP
        `[Jung et al., 2022] <https://arxiv.org/abs/2209.15145>`_,
        a multivalid conformal prediction method that satisfies coverage guarantees conditioned on group membership
        and non-conformity threshold.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)
        self._coverage = None

    def calibrate(
        self,
        scores: Array,
        groups: Optional[Array] = None,
        thresholds: Optional[Array] = None,
        test_groups: Optional[Array] = None,
        test_thresholds: Optional[Array] = None,
        atol: float = 1e-4,
        rtol: float = 1e-6,
        min_prob_b: Union[float, str] = "auto",
        n_buckets: int = 100,
        n_rounds: int = 1000,
        eta: float = 0.1,
        split: float = 0.8,
        bucket_types: Tuple[str, ...] = (">=", "<="),
        patch_type: str = "additive",
        coverage: float = 0.95,
    ) -> Union[Dict, Tuple[Array, Dict]]:
        """
        Calibrate the model by finding a list of patches to the model that bring the calibration error below a
        certain threshold.

        Parameters
        ----------
        scores: Array
            A list of scores :math:`s(x, y)` computed on the calibration data.
            This should be a one-dimensional array of elements between 0 and 1.
        groups: Array
            A list of groups :math:`g(x)` computed on the calibration data.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        thresholds: Optional[Array]
            The initial model evalutions :math:`f(x)` on the calibration data. If not provided, these are set to 0.
        test_groups: Optional[Array]
            A list of groups :math:`g(x)` computed on the test data.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        test_thresholds: Optional[Array]
            The initial model evaluations :math:`f(x)` on the test data. If not provided, these are set to 0.
        atol: float
            Absolute tolerance on the mean squared error.
        rtol: float
            Relative tolerance on the mean squared error.
        min_prob_b: float
            Minimum probability of the conditioning set :math:`B_t` for the patch to be applied.
        n_buckets: int
            The number of buckets used in the algorithm. The smaller the number of buckets, the simpler the model,
            the better its generalization abilities. If not provided, We start from 2 buckets, and progressively double
            the number of buckets until we find a value for which the calibration error falls below the given
            tolerance. Such number of buckets is guaranteed to exist.
        n_rounds: int
            The maximum number of rounds to run the method for.
        eta: float
            Step size. By default, this is set to 1.
        split: float
            Split the calibration data into calibration and validation, according to the given proportion.
            The validation data will be used for early stopping.
        bucket_types: Tuple[str, ...]
            Types of buckets. The following types are currently supported:

            - "=", corresponding of buckets like :math:`\{f(x) = v\}`;
            - ">=", corresponding of buckets like :math:`\{f(x) \ge v\}`;
            - "<=", corresponding of buckets like :math:`\{f(x) \le v\}`.
        coverage: float
            The desired level of coverage. This must be a scalar between 0 and 1.
        Returns
        -------
        Union[Dict, Tuple[Array, Dict]]
            A status including the number of rounds taken to reach convergence and the calibration errors computed
            during the training procedure. if `test_thresholds` and `test_groups` are provided, the list of patches will
            be applied to `test_thresholds`, and the calibrated test thresholds will be returned together with the status.
        """
        self._check_coverage(coverage)
        self._coverage = coverage
        return super().calibrate(
            scores=scores,
            groups=groups,
            values=thresholds,
            test_groups=test_groups,
            test_values=test_thresholds,
            atol=atol,
            rtol=rtol,
            min_prob_b=min_prob_b,
            n_buckets=n_buckets,
            n_rounds=n_rounds,
            eta=eta,
            split=split,
            bucket_types=bucket_types,
            patch_type=patch_type,
            coverage=coverage,
        )

    def apply_patches(
        self,
        groups: Optional[Array] = None,
        thresholds: Optional[Array] = None,
    ) -> Array:
        return super().apply_patches(groups=groups, values=thresholds)

    def calibration_error(
        self,
        scores: Array,
        groups: Optional[Array] = None,
        thresholds: Optional[Array] = None,
        n_buckets: int = 10,
        **kwargs,
    ) -> Array:
        return super().calibration_error(
            scores=scores,
            groups=groups,
            values=thresholds,
            n_buckets=n_buckets,
            coverage=self._coverage,
            **kwargs,
        )

    def _calibration_error(
        self,
        v: Array,
        g: Array,
        c: Array,
        tau: Array,
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        coverage: float = None,
        threshold: Array = None,
    ):
        prob_error, b, prob_b = self._compute_probability_error(
            v=v,
            g=g,
            tau=tau,
            scores=scores,
            groups=groups,
            values=values,
            n_buckets=len(buckets),
            return_prob_b=True,
            coverage=coverage,
            threshold=threshold,
        )
        return prob_b * prob_error, b

    def _compute_probability_error(
        self,
        v: Array,
        g: Array,
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        n_buckets: int,
        return_prob_b: bool = False,
        coverage: float = None,
        threshold: Array = None,
        b: Optional[Array] = None,
    ):
        prob = self._compute_probability(
            v=v,
            g=g,
            tau=tau,
            scores=scores,
            groups=groups,
            values=values,
            n_buckets=n_buckets,
            return_prob_b=return_prob_b,
            threshold=threshold,
            b=b,
        )
        if return_prob_b:
            prob, b, prob_b = prob
            return (coverage - prob) ** 2, b, prob_b
        prob, b = prob
        return (coverage - prob) ** 2, b

    def _compute_probability(
        self,
        v: Array,
        g: Array,
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        n_buckets: int,
        return_prob_b: bool = False,
        threshold: Array = None,
        b: Optional[Array] = None,
    ):
        if b is None:
            b = self._get_b(
                groups=groups,
                values=values,
                v=v,
                g=g,
                tau=tau,
                c=None,
                n_buckets=n_buckets,
            )
        conds = (scores <= (values if threshold is None else threshold)) * b
        prob_b = jnp.mean(b)
        prob = jnp.where(prob_b > 0, jnp.mean(conds) / prob_b, 0.0)
        if return_prob_b:
            return prob, b, prob_b
        return prob, b

    def _get_patch(
        self,
        v: Array,
        g: Array,
        c: Array,
        b: Optional[Array],
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        patch_type: str,
        coverage: float = None,
    ) -> Array:
        if patch_type == "multiplicative":
            raise NotImplementedError("Please use `patch_type='additive'`.")
        n_buckets = len(buckets)
        buckets = jnp.concatenate((-buckets[::-1][:-1], buckets))
        errors, b = vmap(
            lambda delta: self._compute_probability_error(
                v=v,
                g=g,
                tau=None,
                scores=scores,
                groups=groups,
                values=values,
                n_buckets=n_buckets,
                coverage=coverage,
                threshold=values + delta,
                b=b,
            )
        )(buckets)
        min_error = jnp.min(errors)
        indices_min_error = jnp.where(errors == min_error)[0]
        if jnp.sum(indices_min_error >= n_buckets):
            return buckets[jnp.min(indices_min_error)]
        return buckets[jnp.max(indices_min_error)]

    @staticmethod
    def _maybe_check_values(
        values: Optional[Array], test_values: Optional[Array] = None
    ):
        if jnp.any(values < 0) or jnp.any(values > 1):
            raise ValueError("All elements in `thresholds` must be within [0, 1].")

    @staticmethod
    def _check_coverage(coverage: float):
        if coverage < 0 or coverage > 1:
            raise ValueError("`coverage` must be a float between 0 and 1.")



from typing import Callable

from jax import vmap
import jax.numpy as jnp

from fortuna.kernel_regression.kernels.gaussian import gaussian_kernel
from fortuna.typing import Array


class NadarayaWatsonKernelRegressor:
    def __init__(
        self,
        train_inputs: Array,
        train_targets: Array,
        kernel: Callable[[Array], Array] = gaussian_kernel,
    ):
        """
        A Nadaraya-Watson kernel regressor.

        Parameters
        ----------
        kernel: Callable[[Array], Array]
            A kernel.
        train_inputs: Array
            Training inputs.
        train_targets: Array
            Training targets.
        """
        self.kernel = kernel

        if (train_inputs.ndim > 1) or (train_targets.ndim > 1):
            raise ValueError(
                "Both `train_inputs` and `train_targets` must be one-dimensional arrays."
            )
        if len(train_inputs) != len(train_targets):
            raise ValueError(
                "`train_inputs` and `train_targets` must have the same length."
            )
        self.train_inputs = jnp.copy(train_inputs)
        self._mean_train_targets = jnp.mean(train_targets)
        self._std_train_targets = jnp.std(train_targets)
        self.standardized_train_targets = (
            jnp.copy(train_targets) - self._mean_train_targets
        ) / self._std_train_targets

    def predict(self, inputs: Array, bandwidth: float = 0.1):
        """
        Predict the target

        Parameters
        ----------
        inputs: Array
            Inputs.
        bandwidth: float
            Kernel bandwidth.

        Returns
        -------
        Predictions for the given inputs.
        """
        """"""
        if inputs.ndim > 1:
            raise ValueError("`inputs` must be a one-dimensional array.")
        inputs = jnp.copy(inputs)
        kernels = vmap(
            lambda x: self.evaluate_scaled_kernel(
                x - self.train_inputs, bandwidth=bandwidth
            )
        )(inputs)
        m = jnp.sum(kernels * self.standardized_train_targets[None], axis=1) / jnp.sum(
            kernels, axis=1
        )
        m *= self._std_train_targets
        m += self._mean_train_targets
        return m

    def evaluate_scaled_kernel(self, inputs: Array, bandwidth: float) -> Array:
        """
        Given a kernel :math:`K(x)`, Evaluate the scaled kernel :math:`K_h(x) := \frac{1}{h}K\left(\frac{x}{h}\right)`.

        Parameters
        ----------
        inputs: Array
            Inputs.
        bandwidth: Array
            Bandwidth.

        Returns
        -------
        Evaluation of the scaled kernel.
        """
        return self.kernel(inputs / bandwidth) / bandwidth



from jax import vmap

from fortuna.conformal.classification.base import (
    CVPlusConformalClassifier,
    SplitConformalClassifier,
)
from fortuna.typing import Array


@vmap
def _score_fn(probs: Array, target: Array):
    return 1 - probs[target]


def score_fn(
    probs: Array,
    targets: Array,
):
    return _score_fn(probs, targets)


class SimplePredictionConformalClassifier(SplitConformalClassifier):
    def score_fn(
        self,
        probs: Array,
        targets: Array,
    ):
        return score_fn(probs=probs, targets=targets)


class CVPlusSimplePredictionConformalClassifier(CVPlusConformalClassifier):
    def score_fn(
        self,
        probs: Array,
        targets: Array,
    ):
        return score_fn(probs=probs, targets=targets)



import re


def string_cleaner(text: str) -> str:
    """
    Clean a string of text. Remove possible spaces before punctuation and format for proper capitalization.

    Parameters
    ----------
    text: str
        A string of text
    Returns
    -------
    str
        Formatted string.
    """
    text = re.sub(r'\s([?.,%!"](?:\s|$))', r"\1", text)

    text = ". ".join(map(lambda s: s.strip().capitalize(), text.split(".")))
    text = "? ".join(map(lambda s: s.strip().capitalize(), text.split("?")))
    text = "! ".join(map(lambda s: s.strip().capitalize(), text.split("!")))
    text = " ' ".join(map(lambda s: s.strip(), text.split("'")))
    text = text.replace(" ' ", "'")
    return text



"""
Wide ResNet model
(adapted from https://github.com/google/flax/blob/v0.2/examples/cifar10/models/wideresnet.py)
"""
from functools import partial
from typing import (
    Any,
    Callable,
    Tuple,
)

import flax.linen as nn
import jax.numpy as jnp

from fortuna.model.utils.spectral_norm import WithSpectralConv2DNorm
from fortuna.typing import Array

ModuleDef = Any


class WideResnetBlock(nn.Module):
    """
    A wide residual network block.

    Attributes
    ----------
    conv: ModuleDef
        Convolution module.
    norm: ModuleDef
        Normalization module.
    activation: Callable
        Activation function.
    filters: int
        Number of filters.
    strides: Tuple[int, int]
        Strides.
    dropout: ModuleDef
        Dropout module.
    dropout_rate: float
        Dropout rate.
    """

    conv: ModuleDef
    norm: ModuleDef
    activation: Callable
    filters: int
    strides: Tuple[int, int] = (1, 1)
    dropout: ModuleDef = nn.Dropout
    dropout_rate: float = 0.0

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True) -> jnp.ndarray:
        """
        Block forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Block inputs.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Block outputs.
        """
        dropout = self.dropout(rate=self.dropout_rate, broadcast_dims=(1, 2))

        y = self.norm(name="bn1")(x)
        y = nn.relu(y)
        if self.dropout_rate > 0.0:
            y = dropout(y, deterministic=not train)
        y = self.conv(self.filters, (3, 3), self.strides, name="conv1")(y)
        y = self.norm(name="bn2")(y)
        y = nn.relu(y)
        if self.dropout_rate > 0.0:
            y = dropout(y, deterministic=not train)
        y = self.conv(self.filters, (3, 3), name="conv2")(y)

        # Apply an up projection in case of channel mismatch
        if (x.shape[-1] != self.filters) or self.strides != (1, 1):
            x = self.conv(self.filters, (3, 3), self.strides)(x)
        return x + y


class WideResnetGroup(nn.Module):
    """
    A wide residual network group.

    Attributes
    ----------
    conv: ModuleDef
        Convolution module.
    norm: ModuleDef
        Normalization module.
    activation: Callable
        Activation function.
    blocks_per_group: int
        Number of blocks per group.
    strides: Tuple[int, int]
        Strides.
    dropout: ModuleDef
        Dropout module.
    dropout_rate: float
        Dropout rate.
    """

    conv: ModuleDef
    norm: ModuleDef
    activation: Callable
    blocks_per_group: int
    filters: int
    strides: Tuple[int, int] = (1, 1)
    dropout: ModuleDef = nn.Dropout
    dropout_rate: float = 0.0

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True) -> jnp.ndarray:
        """
        Group forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Group inputs.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Group outputs.
        """
        for i in range(self.blocks_per_group):
            x = WideResnetBlock(
                conv=self.conv,
                norm=self.norm,
                activation=self.activation,
                filters=self.filters,
                strides=self.strides if i == 0 else (1, 1),
                dropout=self.dropout,
                dropout_rate=self.dropout_rate,
            )(x, train=train)
        return x


class DeepFeatureExtractorSubNet(nn.Module):
    """
    Deep feature extractor subnetwork.

    Attributes
    ----------
    depth: int
        Depth of the subnetwork.
    widen_factor: int
        Widening factor.
    dropout_rate: float
        Dropout rate.
    dtype: Any
        Layers' dtype.
    activation: Callable
        Activation function.
    conv: ModuleDef
        Convolution module.
    dropout: ModuleDef
        Dropout module.
    """

    depth: int = 28
    widen_factor: int = 10
    dropout_rate: float = 0.0
    dtype: Any = jnp.float32
    activation: Callable = nn.relu
    conv: ModuleDef = nn.Conv
    dropout: ModuleDef = nn.Dropout

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> jnp.ndarray:
        """
        Deep feature extractor subnetwork forward pass.

        Parameters
        ----------
        x: Array
            Input data.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Deep feature extractor representation.
        """
        if hasattr(self, "spectral_norm"):
            conv = self.spectral_norm(self.conv, train=train)
        else:
            conv = self.conv

        blocks_per_group = (self.depth - 4) // 6

        dropout = self.dropout(rate=self.dropout_rate, broadcast_dims=(1, 2))
        conv = partial(conv, use_bias=False, dtype=self.dtype)
        norm = partial(
            nn.BatchNorm,
            use_running_average=not train,
            momentum=0.9,
            epsilon=1e-5,
            dtype=self.dtype,
        )

        x = conv(16, (3, 3), name="init_conv")(x)
        if self.dropout_rate > 0.0:
            x = dropout(x, deterministic=not train)
        x = WideResnetGroup(
            conv=conv,
            norm=norm,
            activation=self.activation,
            blocks_per_group=blocks_per_group,
            filters=16 * self.widen_factor,
            strides=(1, 1),
            dropout=self.dropout,
            dropout_rate=self.dropout_rate,
        )(x, train=train)
        x = WideResnetGroup(
            conv=conv,
            norm=norm,
            activation=self.activation,
            blocks_per_group=blocks_per_group,
            filters=32 * self.widen_factor,
            strides=(2, 2),
            dropout=self.dropout,
            dropout_rate=self.dropout_rate,
        )(x, train=train)
        x = WideResnetGroup(
            conv=conv,
            norm=norm,
            activation=self.activation,
            blocks_per_group=blocks_per_group,
            filters=64 * self.widen_factor,
            strides=(2, 2),
            dropout=self.dropout,
            dropout_rate=self.dropout_rate,
        )(x, train=train)
        x = norm()(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, (8, 8))
        x = x.reshape((x.shape[0], -1))
        return x


class OutputSubNet(nn.Module):
    """
    Output subnetwork.

    Parameters
    ----------
    output_dim: int
        Output dimension.
    dtype: Any
        Layers' dtype.
    """

    output_dim: int
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True) -> jnp.ndarray:
        """
        Output subnetwork forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Subnetwork inputs.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Outputs.
        """
        x = nn.Dense(self.output_dim, dtype=self.dtype)(x)
        return x


class WideResNet(nn.Module):
    """
    Wide residual network class.

    Attributes
    ----------
    output_dim: int
        Output dimension.
    depth: int
        Depth of the subnetwork.
    widen_factor: int
        Widening factor.
    dropout_rate: float
        Dropout rate.
    dtype: Any
        Layers' dtype.
    activation: Callable
        Activation function.
    conv: ModuleDef
        Convolution module.
    dropout: ModuleDef
        Dropout module.
    """

    output_dim: int
    depth: int = 28
    widen_factor: int = 10
    dropout_rate: float = 0.0
    dtype: Any = jnp.float32
    activation: Callable = nn.relu
    conv: ModuleDef = nn.Conv
    dropout: ModuleDef = nn.Dropout

    def setup(self):
        self.dfe_subnet = DeepFeatureExtractorSubNet(
            depth=self.depth,
            widen_factor=self.widen_factor,
            dropout_rate=self.dropout_rate,
            dtype=self.dtype,
            activation=self.activation,
            conv=self.conv,
            dropout=self.dropout,
        )
        self.output_subnet = OutputSubNet(output_dim=self.output_dim, dtype=self.dtype)

    def __call__(self, x: Array, train: bool = True) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Input data.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Outputs.
        """
        x = self.dfe_subnet(x, train)
        x = self.output_subnet(x, train)
        return x


WideResNet28_10 = partial(WideResNet, depth=28, widen_factor=10)


class WideResNetDeepFeatureExtractorSubNetWithSN(
    WithSpectralConv2DNorm, DeepFeatureExtractorSubNet
):
    pass


# define the feature extractors with spectral norm
WideResNetD28W10DeepFeatureExtractorSubNetWithSN = partial(
    WideResNetDeepFeatureExtractorSubNetWithSN, depth=28, widen_factor=10
)



from typing import (
    Dict,
    Iterable,
    Optional,
    Tuple,
    Union,
)

from jax import numpy as jnp

from fortuna.data.loader.base import (
    BaseDataLoaderABC,
    BaseInputsLoader,
    BaseTargetsLoader,
)
from fortuna.data.loader.utils import IterableData
from fortuna.typing import (
    Array,
    Shape,
)


class HuggingFaceDataLoader(BaseDataLoaderABC):
    def __init__(
        self,
        iterable: Union[
            Iterable[Dict[str, Array]], Iterable[Tuple[Dict[str, Array], Array]]
        ],
        num_unique_labels: int = None,
        num_inputs: Optional[int] = None,
    ):
        """
        A data loader class.

        Parameters
        ----------
        iterable : Union[Iterable[Dict[str, Array]], Iterable[Tuple[Dict[str, Array],Array]]]
            A data loader obtained via :func:`~HuggingFaceClassificationDataset.get_dataloader`.
        num_unique_labels: int
            Number of unique target labels in the task (classification only)
        num_inputs: Optional[int]
            Number of data points.
        """
        super().__init__(iterable, num_unique_labels)
        self._num_inputs = num_inputs

    @property
    def size(self):
        if self._num_inputs:
            return self._num_inputs
        else:
            return super().size

    @property
    def input_shape(self) -> Shape:
        def fun():
            for inputs, _ in self:
                input_shape = {k: v.shape[1:] for k, v in inputs.items()}
                break
            return input_shape

        return fun()

    @property
    def num_unique_labels(self) -> Optional[int]:
        if self._num_unique_labels is None:
            self._num_unique_labels = len(jnp.unique(self.to_array_targets()))
        return self._num_unique_labels

    def to_array_targets(self):
        targets = []
        for _, batch_targets in self:
            targets.append(batch_targets)
        return jnp.concatenate(targets, 0)

    def to_inputs_loader(self):
        return BaseInputsLoader(IterableData.data_loader_to_inputs_iterable(self))

    def to_targets_loader(self):
        return BaseTargetsLoader(IterableData.data_loader_to_targets_iterable(self))



from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict

from fortuna.prob_model.posterior.posterior_multi_state_repository import (
    PosteriorMultiStateRepository,
)
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    AnyKey,
    OptaxOptimizer,
    Params,
    Path,
)
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)


class SGMCMCPosteriorStateRepository(PosteriorMultiStateRepository):
    def __init__(
        self,
        size: int,
        checkpoint_dir: Optional[Path] = None,
        all_params: Optional[Params] = None,
        which_params: Optional[Tuple[List[AnyKey], ...]] = None,
    ):
        super().__init__(size=size, checkpoint_dir=checkpoint_dir)
        self._all_params = all_params
        self._which_params = which_params

    def get(
        self,
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Union[List[PosteriorState], PosteriorState]:
        state = super().get(
            i=i,
            checkpoint_path=checkpoint_path,
            optimizer=optimizer,
            prefix=prefix,
            **kwargs,
        )
        return self._update_state(state, modify="add")

    def put(
        self,
        state: PosteriorState,
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        keep: int = 1,
        prefix: str = "checkpoint_",
    ) -> None:
        state = self._update_state(state, modify="remove")
        return super().put(
            state=state,
            i=i,
            checkpoint_path=checkpoint_path,
            keep=keep,
            prefix=prefix,
        )

    def pull(
        self,
        i: int = None,
        checkpoint_path: Path = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> PosteriorState:
        state = super().pull(
            i=i,
            checkpoint_path=checkpoint_path,
            optimizer=optimizer,
            prefix=prefix,
            **kwargs,
        )
        return self._update_state(state, modify="add")

    def extract(
        self,
        keys: List[str],
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Union[Dict, List[Dict]]:
        def _extract(_i):
            state = self.get(
                i=_i,
                checkpoint_path=checkpoint_path,
                prefix=prefix,
            )
            return {k: getattr(state, k) for k in keys}

        if i is not None:
            return _extract(i)
        dicts = []
        for i in range(self.size):
            dicts.append(_extract(i))
        return dicts

    def _update_state(
        self,
        state: Union[List[PosteriorState], PosteriorState],
        modify: str,
    ) -> Union[List[PosteriorState], PosteriorState]:
        if self._which_params is None:
            return state

        if isinstance(state, list):
            return [_update_state(_state, modify=modify) for _state in state]

        if modify == "add":
            state = state.replace(
                params=FrozenDict(
                    nested_set(
                        d=self._all_params.unfreeze(),
                        key_paths=self._which_params,
                        objs=tuple(
                            [
                                nested_get(d=state.params, keys=path)
                                for path in self._which_params
                            ]
                        ),
                    )
                )
            )
        elif modify == "remove":
            state = state.replace(
                params=FrozenDict(
                    nested_set(
                        d={},
                        key_paths=self._which_params,
                        objs=tuple(
                            [
                                nested_get(d=state.params, keys=path)
                                for path in self._which_params
                            ]
                        ),
                        allow_nonexistent=True,
                    )
                ),
                step=state.step,
            )
        else:
            raise RuntimeError(f"Invalid update state method {method}.")

        return state



from typing import (
    List,
    Optional,
    Union,
)

from jax import (
    random,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.prob_output_layer.base import ProbOutputLayer
from fortuna.typing import Array


class RegressionProbOutputLayer(ProbOutputLayer):
    def __init__(self):
        r"""
        Regression probabilistic output layers class. It characterizes the probability distribution of a target
        variable given a calibrated output as a Gaussian distribution. That is
        :math:`p(y|\mu, \sigma^2)=\text{Normal}(y|\mu, \sigma^2)`, where :math:`y` denotes a target variable and
        :math:`\omega=[\mu, \log\sigma^2]` a calibrated output.

        """
        super().__init__()
        self.log2pi = jnp.log(2 * jnp.pi)

    def log_prob(self, outputs: Array, targets: Array, **kwargs) -> jnp.ndarray:
        means, log_vars = jnp.split(outputs, 2, axis=-1)
        return -0.5 * jnp.sum(
            jnp.exp(-log_vars) * (targets - means) ** 2 + log_vars + self.log2pi, -1
        )

    def predict(self, outputs: Array, **kwargs) -> jnp.ndarray:
        return jnp.split(outputs, 2, axis=-1)[0]

    def sample(
        self,
        n_target_samples: int,
        outputs: Array,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        means, log_vars = jnp.split(outputs, 2, axis=-1)
        return means + jnp.exp(0.5 * log_vars) * random.normal(
            rng, (n_target_samples,) + means.shape
        )

    def quantile(
        self,
        q: Union[float, Array, List],
        outputs: Array,
        n_target_samples: Optional[int] = 30,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if type(q) == list:
            q = jnp.array(q)
        samples = self.sample(
            n_target_samples=n_target_samples, outputs=outputs, rng=rng
        )
        return jnp.quantile(samples, q, axis=0)

    def mean(self, outputs: Array, **kwargs) -> jnp.ndarray:
        return jnp.split(outputs, 2, axis=-1)[0]

    def mode(self, outputs: Array, **kwargs) -> jnp.ndarray:
        return jnp.split(outputs, 2, axis=-1)[0]

    def variance(self, outputs: Array, **kwargs) -> jnp.ndarray:
        return jnp.exp(jnp.split(outputs, 2, axis=-1)[1])

    def entropy(
        self,
        outputs: Array,
        n_target_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> jnp.ndarray:
        samples = self.sample(n_target_samples, outputs, rng=rng, **kwargs)

        @vmap
        def _log_lik_fun(sample: jnp.ndarray):
            return self.log_prob(outputs, sample, **kwargs)

        return -jnp.mean(_log_lik_fun(samples), 0)

    def credible_interval(
        self,
        outputs: Array,
        n_target_samples: int = 30,
        error: float = 0.05,
        interval_type: str = "two-tailed",
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        r"""
        Estimate credible intervals for the target variable. This is supported only if the target variable is scalar.

        Parameters
        ----------
        outputs: Array
            Model outputs.
        n_target_samples: int
            Number of target samples to draw for each output.
        error: float
            The interval error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of credibility.
        interval_type: str
            The interval type. We support "two-tailed" (default), "right-tailed" and "left-tailed".
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        jnp.ndarray
            A credibility interval for each of the outputs.
        """
        supported_types = ["two-tailed", "right-tailed", "left-tailed"]
        if interval_type not in supported_types:
            raise ValueError(
                "`type={}` not recognised. Please choose among the following supported types: {}.".format(
                    supported_types
                )
            )
        q = (
            jnp.array([0.5 * error, 1 - 0.5 * error])
            if interval_type == "two-tailed"
            else error
            if interval_type == "left-tailed"
            else 1 - error
        )
        qq = self.quantile(
            q=q, outputs=outputs, n_target_samples=n_target_samples, rng=rng
        )
        if qq.shape[-1] != 1:
            raise ValueError(
                """Credibility intervals are only supported for scalar target variables."""
            )
        if interval_type == "two-tailed":
            lq, uq = qq.squeeze(2)
            return jnp.array(list(zip(lq, uq)))
        else:
            return qq



import logging
import pickle
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from sklearn.mixture import GaussianMixture
import torch
from torch import nn
from tqdm import tqdm
from transformers import PreTrainedTokenizer
import umap.umap_ as umap

from fortuna.conformal import BinaryClassificationMulticalibrator
from fortuna.hallucination.grouping.clustering.base import GroupingModel
from fortuna.hallucination.scoring.inv_perplexity import inv_perplexity


class HallucinationMulticalibrator:
    def __init__(
        self,
        generative_model: nn.Module,
        tokenizer: PreTrainedTokenizer,
        embedding_reduction_model: Optional = None,
        clustering_models: Optional[List] = None,
        scoring_fn: Optional[
            Callable[[torch.Tensor, torch.Tensor, int], torch.Tensor]
        ] = None,
        seed: int = 0,
    ):
        """
        A hallucination multicalibrator class.
        Given some context (e.g. a question), some text (e.g. one or multiple answers) and some scoring function,
        it multicalibrates the probability that the text is a hallucination or not based on a supervised calibration
        dataset.
        Under the hood, the method constructs model-based low-dimensional embeddings of the initial contexts plus texts,
        it clusters them to provide meaningful (possibly overlapping) groups,
        it computes the scores for each text,
        and it finally calibrates the scores on each subset formed by the intersection of each group and each level set
        of the score function.

        Parameters
        ----------
        generative_model: nn.Module
            A generative model.
        tokenizer: PreTrainedTokenizer
            A tokenizer.
        embedding_reduction_model: Optional
            An embedding reduction model.
        clustering_models: Optional[List]
            A list of clustering models.
        scoring_fn: Optional[Callable[[torch.Tensor, torch.Tensor, int], torch.Tensor]]
            A scoring function.
        """
        self.generative_model = generative_model
        self.tokenizer = tokenizer
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token
            logging.info("`tokenizer.pad_token` is None. Set to `tokenizer.eos_token`.")
        self.embedding_reduction_model = embedding_reduction_model or umap.UMAP(
            n_neighbors=100, n_components=100
        )
        self.scoring_fn = scoring_fn or inv_perplexity
        self.clustering_models = clustering_models or [
            GaussianMixture(n_components=i) for i in range(2, 21)
        ]
        self.grouping_model = None
        self.multicalibrator = None
        self._quantiles = None
        self.rng = np.random.default_rng(seed)

    def fit(
        self,
        texts: List[str],
        contexts: List[str],
        targets: List[int],
        batch_size: int = 16,
        quantile_group_scores_threshold: float = 0.8,
        balance: bool = False,
    ) -> Dict:
        """
        Fit the multicalibrator.
        Internally, it first embeds the inputs using the generative model.
        Then it reduces the embeddings dimensionality.
        Then it fits the provided clustering models and finds the best according to the BIC.
        Finally, it computes the scores and applies the multicalibrator.

        Parameters
        ----------
        texts: Union[List[str], List[List[str]]]
            The texts to fit.
            This may either be a list of strings (e.g. a list of single answers),
            or a list of lists of strings (e.g. a list of multi-choice answers).
        contexts: List[str]
            A list of contexts (e.g. a list of questions).
        targets: Union[List[str]]
            A list of target variables to be used for calibration.
            If `texts` is a list of strings, `targets` should be binary variables indicating whether each of the strings
            in the `texts` list should be marked as positive given the corresponding `contexts`.
        batch_size: int
            The batch size.
        quantile_group_scores_threshold: float
            A threshold for which to compute the quantiles of the clustering scores.
            This will determine the groups.
        balance: bool
            Whether to balance the calibration data.

        Returns
        -------
        Dict
            The status returned by fitting the multicalibrator.
        """
        if balance:
            texts, contexts, targets = self._balance_data(texts, contexts, targets)

        (
            scores,
            embeddings,
        ) = self._compute_scores_embeddings(
            texts=texts, contexts=contexts, batch_size=batch_size
        )
        targets = np.array(targets, dtype="int32")

        embeddings = self.embedding_reduction_model.fit_transform(embeddings)
        embeddings = np.concatenate((embeddings, scores[:, None]), axis=1)

        self.grouping_model = GroupingModel()
        self.grouping_model.fit(embeddings, clustering_models=self.clustering_models)

        group_scores = self.grouping_model.predict_proba(
            embeddings=embeddings,
        )
        self._quantiles = self.grouping_model.compute_quantile(
            probs=group_scores, threshold=quantile_group_scores_threshold
        )
        groups = self._get_groups(group_scores)

        self.multicalibrator = BinaryClassificationMulticalibrator()
        status = self.multicalibrator.calibrate(
            probs=scores, targets=targets, groups=groups
        )

        return status

    def predict_proba(
        self,
        texts: List[str],
        contexts: List[str],
        batch_size: int = 16,
        calibrate: bool = True,
    ) -> np.ndarray:
        """
        Predict probabilities of positive classes for each text given each context.

        Parameters
        ----------
        texts: List[str]
            The texts to fit.
            This may either be a list of strings (e.g. a list of single answers),
            or a list of lists of strings (e.g. a list of multi-choice answers).
        contexts: List[str]
            A list of contexts (e.g. a list of questions).
        batch_size: int
            The batch size.
        calibrate: bool
            Whether to calibration the initial probability estimates.

        Returns
        -------
        np.ndarray
            The predicted probabilities.
        """
        if self.multicalibrator is None:
            raise ValueError("`fit` must be called before this method.")

        (scores, embeddings) = self._compute_scores_embeddings(
            texts=texts, contexts=contexts, batch_size=batch_size
        )

        if not calibrate:
            return scores

        embeddings = self.embedding_reduction_model.transform(embeddings)
        embeddings = np.concatenate((embeddings, scores[:, None]), axis=1)

        group_scores = self.grouping_model.predict_proba(
            embeddings=embeddings,
        )
        groups = self._get_groups(group_scores)

        return self.multicalibrator.apply_patches(probs=scores, groups=groups)

    def predict(
        self,
        texts: List[str],
        contexts: List[str],
        batch_size: int = 16,
        calibrate: bool = True,
        probs: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        A binary prediction for each text given each context.
        The method returns 1 of a predicted is greater than 0.5, and 0 otherwise.

        Parameters
        ----------
        texts: List[str],
            The texts to fit.
            This may either be a list of strings (e.g. a list of single answers),
            or a list of lists of strings (e.g. a list of multi-choice answers).
        contexts: List[str]
            A list of contexts (e.g. a list of questions).
        batch_size: int
            The batch size.
        calibrate: bool
            Whether to calibration the initial probability estimates.
        probs: Optional[np.ndarray]
            Predicted probabilities. If these are available, the method does not compute them under the hood.

        Returns
        -------
        np.ndarray
            The predicted probabilities.
        """
        if probs is None:
            probs = self.predict_proba(
                texts=texts,
                contexts=contexts,
                batch_size=batch_size,
                calibrate=calibrate,
            )
        return (probs >= 0.5).astype(int)

    def _get_groups(self, group_scores: np.ndarray):
        groups = group_scores >= self._quantiles[None]
        return np.concatenate((groups, np.ones((groups.shape[0], 1))), axis=1).astype(
            bool
        )

    def _compute_scores_embeddings(
        self, texts: List[str], contexts: List[str], batch_size: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        scores = []
        embeddings = []

        gen = self._batch(texts, contexts, batch_size)

        for batch_texts, batch_contexts in tqdm(
            gen, total=int(np.ceil(len(texts) / batch_size))
        ):
            logits, _scores = self._get_logits_scores(batch_texts, batch_contexts)
            embeddings.append(logits.mean(1))
            scores.append(_scores)

        return (
            np.concatenate(scores, axis=0).astype("float32"),
            np.concatenate(embeddings, axis=0),
        )

    @staticmethod
    def _batch(texts: List[str], contexts: List[str], batch_size: int):
        for i in range(0, len(texts), batch_size):
            yield texts[i : i + batch_size], contexts[i : i + batch_size]

    def _get_logits_scores(
        self, texts: str, contexts: str
    ) -> Tuple[np.ndarray, np.ndarray]:
        context_inputs = self.tokenizer(contexts, return_tensors="pt", padding=True).to(
            self.generative_model.device
        )
        text_inputs = self.tokenizer(texts, return_tensors="pt", padding=True).to(
            self.generative_model.device
        )
        inputs = {
            k: torch.cat((context_inputs[k], v), dim=1) for k, v in text_inputs.items()
        }

        with torch.no_grad():
            _logits = self.generative_model(**inputs).logits

            _scores = self.scoring_fn(
                logits=_logits,
                labels=inputs["input_ids"],
                n_final_tokens=text_inputs.input_ids.shape[1],
            )

        return _logits.cpu().numpy(), _scores.cpu().numpy()

    def _balance_data(
        self, texts: List[str], contexts: List[str], targets: List[int]
    ) -> Tuple[List[str], List[str], List[int]]:
        idx0 = [i for i, y in enumerate(targets) if y == 0]
        idx1 = [i for i, y in enumerate(targets) if y == 1]
        len_diff = len(idx1) - len(idx0)
        idx = self.rng.choice(
            idx0 if len_diff > 0 else idx1, np.abs(len_diff), replace=True
        )
        for i in idx:
            texts.append(texts[i])
            contexts.append(contexts[i])
            targets.append(targets[i])
        return texts, contexts, targets

    def save(self, path):
        state = dict(
            embedding_reduction_model=self.embedding_reduction_model,
            grouping_model=self.grouping_model,
            multicalibrator=self.multicalibrator,
            _quantiles=self._quantiles,
        )

        with open(path, "wb") as filehandler:
            pickle.dump(state, filehandler, -1)

    def load(self, path):
        state = pickle.load(open(path, "rb"))
        for k, v in state.items():
            setattr(self, k, v)



from typing import Optional

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.data.loader import (
    DataLoader,
    InputsLoader,
)
from fortuna.likelihood.base import Likelihood
from fortuna.utils.random import WithRNG


class Predictive(WithRNG):
    def __init__(self, likelihood: Likelihood):
        """
        Predictive distribution abstract class.

        Parameters
        ----------
        likelihood : Likelihood
             A likelihood object.
        """
        self.likelihood = likelihood
        self.state = None

    def log_prob(self, data_loader: DataLoader, distribute: bool = True) -> jnp.ndarray:
        r"""
        Estimate the predictive log-probability density function (a.k.a. log-pdf), that is

        .. math::
            \log p(y|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`y` is an observed target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive log-pdf for each data point.
        """
        state = self.state.get()
        return self.likelihood.log_prob(
            params=state.params,
            data_loader=data_loader,
            mutable=state.mutable,
            distribute=distribute,
        )

    def sample(
        self,
        inputs_loader: InputsLoader,
        n_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Sample from an approximation of the predictive distribution for each input data point, that is

        .. math::
            y^{(i)}\sim p(\cdot|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`y^{(i)}` is a sample of the target variable for the input :math:`x`.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_samples : int
            Number of target samples to sample for each input data point.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            Samples for each input data point.
        """
        state = self.state.get()
        return self.likelihood.sample(
            n_target_samples=n_samples,
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            rng=rng,
            distribute=distribute,
        )

    def mean(
        self,
        inputs_loader: InputsLoader,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mean of the target variable, that is

        .. math::
            \mathbb{E}_{Y|x, \mathcal{D}}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mean for each input.
        """
        state = self.state.get()
        return self.likelihood.mean(
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            distribute=distribute,
        )

    def mode(
        self,
        inputs_loader: InputsLoader,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mode of the target variable, that is

        .. math::
            \text{argmax}_y\ p(y|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`y` is the target variable to optimize upon.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mode for each input.
        """
        state = self.state.get()
        return self.likelihood.mode(
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            distribute=distribute,
        )

    def variance(
        self,
        inputs_loader: InputsLoader,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive variance of the target variable, that is

        .. math::
            \text{Var}_{Y|x, D}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Note that the predictive variance above corresponds to the sum of its aleatoric and epistemic components.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive variance for each input.
        """
        state = self.state.get()
        return self.likelihood.variance(
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            distribute=distribute,
        )

    def std(
        self,
        inputs_loader: InputsLoader,
        variances: Optional[jnp.ndarray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive standard deviation of the target variable, that is

        .. math::
            \text{Var}_{Y|x, D}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        variances: Optional[jnp.ndarray]
            An estimate of the predictive variance.passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive standard deviation for each input.
        """
        if variances is None:
            variances = self.variance(
                inputs_loader=inputs_loader,
                distribute=distribute,
            )
        return jnp.sqrt(variances)



from fortuna.data.loader.array_loaders import (
    DataLoader,
    InputsLoader,
    TargetsLoader,
)
from fortuna.data.loader.base import (
    BaseInputsLoader,
    BaseTargetsLoader,
    ConcatenatedLoader,
    DeviceDimensionAugmentedLoader,
)



from typing import (
    Optional,
    Union,
)

import jax.numpy as jnp

from fortuna.conformal.multivalid.mixins.classification.top_label_multicalibrator import (
    TopLabelMulticalibratorMixin,
)
from fortuna.conformal.multivalid.one_shot.multicalibrator import OneShotMulticalibrator
from fortuna.typing import Array


class OneShotTopLabelMulticalibrator(
    TopLabelMulticalibratorMixin, OneShotMulticalibrator
):
    def __init__(self, n_classes: int, seed: int = 0):
        super().__init__(n_classes=n_classes, seed=seed)

    def calibrate(
        self,
        targets: Array,
        probs: Optional[Array] = None,
        test_probs: Optional[Array] = None,
        n_buckets: int = 100,
        min_prob_b: Union[float, str] = "auto",
    ):
        return super().calibrate(
            scores=self._get_scores(targets),
            values=probs,
            test_values=test_probs,
            n_buckets=n_buckets,
            min_prob_b=min_prob_b,
        )

    def apply_patches(
        self,
        probs: Array,
    ) -> Array:
        return super().apply_patches(values=probs)

    @staticmethod
    def _get_b(
        values: Array,
        v: Array,
        c: Optional[Array],
        n_buckets: int,
    ) -> Array:
        return (jnp.abs(values[:, c] - v) < 0.5 / n_buckets) * (values.argmax(1) == c)



import inspect
from typing import (
    List,
    Optional,
)

import jax.numpy as jnp

from fortuna.conformal.classification.base import ConformalClassifier
from fortuna.typing import Array


class AdaptiveConformalClassifier(ConformalClassifier):
    def __init__(self, conformal_classifier: ConformalClassifier):
        """
        An adaptive conformal classifier class
        (see `Gibbs & Candes, 2021 <https://proceedings.neurips.cc/paper/2021/hash/0d441de75945e5acbc865406fc9a2559-Abstract.html>`_).
        It takes any conformal classifier and adds the functionality to update the coverage error to take into account
        distributional shifts in the data.

        Parameters
        ----------
        conformal_classifier: ConformalClassifier
            A conformal method for classification.
        """
        for s, m in inspect.getmembers(conformal_classifier):
            if not s.startswith("__"):
                setattr(self, s, m)

    def update_error(
        self,
        conformal_set: List[int],
        error: float,
        target: Array,
        target_error: float,
        gamma: float = 0.005,
        weights: Optional[Array] = None,
        were_in: Optional[Array] = None,
        return_were_in: bool = False,
    ) -> Array:
        """
        Update the coverage error based on the test target variable belonging or not to the conformal set.

        Parameters
        ----------
        conformal_set: List[int]
            A conformal set for the current test target variable.
        error: float
            The current coverage error to update.
        target: Array
            The observed test target variable.
        target_error: float
            The target coverage error.
        gamma: float
            The step size for the coverage error update.
        weights: Optional[Array]
            Weights over the considered past time steps and the current one.
            This must be a one-dimensional array of increasing components between 0 and 1, summing up to 1.
        were_in: Optional[Array]
            It indicates whether the target variables of the considered past time steps fell within the respective
            conformal sets. This must be a one-dimensional array of 1's and 0's. Its length must be the length of
            `weights` minus one, as it refers to all the past time steps but not the current one.
        return_were_in: bool
            It returns an updated `were_in`, which includes whether the current test target variable falls within its
            conformal set.

        Returns
        -------
        Array
            The updated coverage error.
        """
        if gamma <= 0:
            raise ValueError(
                f"`gamma` must be a value greater than 0, but {gamma} was found."
            )
        if weights is not None and were_in is None:
            raise ValueError(
                "If `weights` is available, `were_in` must be available too."
            )
        if weights is None and were_in is not None:
            raise ValueError(
                "If `were_in` is available, `weights` must be available too."
            )
        if weights is not None:
            if weights.ndim > 1:
                raise ValueError(
                    "`weights` must be a one-dimensional array over the considered times in the time "
                    "series."
                )
            if (
                jnp.any(weights[:-1] > weights[1:])
                or jnp.any(weights < 0)
                or jnp.any(weights > 1)
                or not jnp.allclose(jnp.sum(weights), 1.0)
            ):
                raise ValueError(
                    "`weights` must be a vector of weights sorted in ascending order, with all "
                    "elements between 0 and 1, summing up to 1."
                )
        if were_in is not None:
            if jnp.any((were_in != 0) * (were_in != 1)):
                raise ValueError("`were_in` must be a vector of 0's and 1's.")
            if were_in.ndim != 1:
                raise ValueError(
                    "`were_in` must a be one-dimensional array over the considered times in the time "
                    "series."
                )
            if len(were_in) != len(weights) - 1:
                raise ValueError(
                    "`len(weights)-1` and `len(were_in)` must be the same. "
                    f"However, {len(weights) - 1} and {len(were_in)} were found, respectively."
                )
        is_in = self.is_in(target[None], [conformal_set])
        if were_in is not None:
            is_in = jnp.concatenate((were_in, is_in))
            error += gamma * (target_error - jnp.dot(weights, 1 - is_in))
        else:
            error += gamma * (target_error - 1 + is_in.squeeze())
        if error > 1:
            error = 1
        if error < 0:
            error = 0

        if return_were_in:
            return float(error), is_in
        return error



import logging
import pathlib

import boto3
from hydra import (
    compose,
    initialize_config_dir,
)
from hydra.utils import instantiate
import sagemaker
from sagemaker.debugger import (
    FrameworkProfile,
    ProfilerConfig,
    ProfilerRule,
    rule_configs,
)
from sagemaker.tensorflow import TensorFlow
from sagemaker.tuner import HyperparameterTuner

import fortuna
from fortuna.sagemaker.utils import (
    DictConfig,
    create_input_channels,
    get_base_job_name,
    get_hparams,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _run_training_job(cfg: DictConfig) -> None:
    """
    Run a training job on Amazon SageMaker.
    The model, the hyperparameters, and all the sagemaker-related parameters
    (image_uri to be used, the model training entry point, the AWS IAM role to use, the instance type/count to use,
    the input/output paths on s3 or local disk) have to be defined in the configuration dictionary.

    Parameters
    ----------
    cfg: DictConfig
        Configuration dictionary.
    """
    if hasattr(cfg.sagemaker, "metrics"):
        metrics = [dict(d) for d in cfg.sagemaker.metrics]
    else:
        metrics = []

    # define the dependency path
    source_dir = pathlib.Path(fortuna.__file__).parents[1]
    # get proper datasets
    channels = create_input_channels(cfg)
    # get entrypoint args
    hparams = get_hparams(cfg=cfg)
    # update channels
    if "model_path" in cfg.model:
        channels.update({"loadmodel": cfg.model.model_path})

    logger.info(f"Training job with parameters: {hparams}")
    logger.info(f"\nEstimating with channels: {channels}")

    base_job_name = get_base_job_name(cfg)

    logger.info(f"Source dir: {source_dir}")

    rules = [
        ProfilerRule.sagemaker(rule_configs.LowGPUUtilization()),
        ProfilerRule.sagemaker(rule_configs.ProfilerReport()),
    ]
    profiler_config = ProfilerConfig(
        system_monitor_interval_millis=500,
        framework_profile_params=FrameworkProfile(
            local_path="/opt/ml/output/profiler/", start_step=5, num_steps=2
        ),
    )

    sm_role = f"arn:aws:iam::{cfg.sagemaker.account_id}:role/{cfg.sagemaker.iam_role}"
    boto_sess = boto3.Session(
        profile_name=cfg.sagemaker.profile, region_name=cfg.sagemaker.region
    )
    sm_sess = sagemaker.Session(boto_session=boto_sess)
    sm = boto_sess.client(service_name="sagemaker")

    estimator = TensorFlow(
        image_uri=f"{cfg.sagemaker.account_id}.dkr.ecr.{cfg.sagemaker.region}.amazonaws.com/fortuna:dev",
        entry_point=f"./{cfg.sagemaker.entrypoint}",
        source_dir=str(source_dir),
        role=sm_role,
        instance_type=cfg.sagemaker.instance_type,
        instance_count=1,
        hyperparameters=hparams,
        output_path=cfg.output_data_path,
        base_job_name=base_job_name,
        metric_definitions=metrics,
        py_version="py38",
        sagemaker_session=sm_sess,
        profiler_config=profiler_config,
        rules=rules,
    )

    if "tuner" in cfg.hyperparams:
        logger.info(f"Starting hyperparams optimization: {cfg.hyperparams.tuner}")
        estimator = HyperparameterTuner(
            estimator=estimator,
            base_tuning_job_name=base_job_name,
            metric_definitions=metrics,
            **instantiate(cfg.hyperparams.tuner, _convert_="partial"),
        )

    estimator.fit(inputs=channels, wait=False)


def run_training_job(config_dir: str, config_filename: str):
    """
    Run a training job on Amazon SageMaker starting from a path to the configuration directory, and the name of the main configuration file.
    All configuration files should be written in `yaml` format. The following information must be provided.

    .. code-block:: yaml

        sagemaker:
          account_id: ~
          iam_role: ~
          entrypoint: ~
          instance_type: ~
          profile: ~
          region: ~

        dataset:
          base_data_path: ~

        task:
          name: ~

        model:
          name: ~

        method:
          name: ~

    `sagemaker` may optionally include `metrics`, which specifies the `training metrics <https://docs.aws.amazon.com/sagemaker/latest/dg/training-metrics.html>`_ to be tracked by SageMaker,
    and `tuner`, which configures the `hyperparameter tuning <https://sagemaker.readthedocs.io/en/stable/api/training/tuner.html>`_.
    Any argument to pass to the entrypoint script must be specified under `hparams` in either `task`, `model` or `method`.
    For example:

    .. code-block:: yaml

        model:
          name: ~
            hparams:
              some_hyperparameter: ~

    For the training job to run successfully, you must have already built and pushed a Docker imagine to AWS in the
    region specified by the configuration file. You can do so by running `bash fortuna/docker/build_and_push.sh`,
    providing the required information as argument. Docker and related plug-ins may need to be installed beforehand.

    Parameters
    ----------
    config_dir: str
        Absolute path to main configuration directory.
    config_filename: str
        Main configuration filename. This should not include the `.yaml` extension.
    """
    with initialize_config_dir(config_dir=config_dir):
        cfg = compose(config_name=config_filename)
    _run_training_job(cfg)



import logging
from typing import (
    Optional,
    Tuple,
)

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.posterior.map.map_approximator import MAPPosteriorApproximator
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.typing import Status
from fortuna.utils.random import RandomNumberGenerator


def run_preliminary_map(
    joint: Joint,
    train_data_loader: DataLoader,
    val_data_loader: DataLoader,
    map_fit_config: Optional[FitConfig],
    rng: RandomNumberGenerator,
    **kwargs,
) -> Tuple[MAPState, Status]:
    logging.info("Do a preliminary run of MAP.")
    map_posterior = MAPPosterior(
        joint, posterior_approximator=MAPPosteriorApproximator()
    )
    map_posterior.rng = rng
    status = map_posterior.fit(
        rng=rng.get(),
        train_data_loader=train_data_loader,
        val_data_loader=val_data_loader,
        fit_config=map_fit_config,
        **kwargs,
    )
    logging.info("Preliminary run with MAP completed.")
    return map_posterior.state.get(), status



from typing import (
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
import flax.linen as nn
from flax.training.checkpoints import PyTree
from jax import random
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.typing import (
    Array,
    CalibMutable,
    CalibParams,
)
from fortuna.utils.random import WithRNG


class OutputCalibManager(WithRNG):
    def __init__(self, output_calibrator: Optional[nn.Module] = None):
        self.output_calibrator = output_calibrator

    def apply(
        self,
        params: CalibParams,
        outputs: Array,
        mutable: Optional[CalibMutable] = None,
        calib: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        """
        Apply the output calibrator forward pass.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        outputs : Array
            Outputs for each data point.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib : bool
            Whether the method is called during calibration.
        rng: Optional[PRNGKeyArray]
            A random number generator.
            If not passed,
            this will be taken from the attributes of this class.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]
            The output of the model manager for each input. Mutable objects may also be returned.
        """
        if self.output_calibrator is None:
            return outputs
        if mutable is None:
            mutable = False
        variables = params.unfreeze()

        # setup dropout key
        if rng is not None:
            rng, dropout_rng = random.split(rng, 2)
            rngs = {"dropout": dropout_rng}
        else:
            rngs = None

        if mutable:
            mutable_variables = mutable.unfreeze()
            variables.update(mutable_variables)
            mutable = list(mutable_variables.keys())
        if calib and mutable:
            outputs, mutable = self.output_calibrator.apply(
                variables, outputs, mutable=mutable, train=calib, rngs=rngs
            )
            return outputs, {"mutable": mutable}
        else:
            return self.output_calibrator.apply(
                variables, outputs, train=calib, mutable=False, rngs=rngs
            )

    def init(
        self, output_dim: int, rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Optional[FrozenDict]:
        """
        Initialize random parameters and mutable objects.

        Parameters
        ----------
        output_dim: int
            The output dimension.
        rng: Optional[PRNGKeyArray]
            A random number generator.
            If not passed,
            this will be taken from the attributes of this class.

        Returns
        -------
        Optional[FrozenDict]
            Initialized random parameters and mutable objects.
        """
        if rng is None:
            rng = self.rng.get()
        rng, params_key, dropout_key = random.split(rng, 3)
        rngs = {"params": params_key, "dropout": dropout_key}
        return (
            self.output_calibrator.init(rngs, jnp.zeros((1, output_dim)), **kwargs)
            if self.output_calibrator is not None
            else None
        )



from __future__ import annotations

from typing import (
    Any,
    Optional,
)

from flax.training import (
    dynamic_scale,
    train_state,
)
import jax.numpy as jnp

from fortuna.typing import Params
from fortuna.utils.strings import convert_string_to_tuple


class TrainState(train_state.TrainState):
    encoded_name: tuple = convert_string_to_tuple("TrainState")
    frozen_params: Optional[Params] = None
    dynamic_scale: Optional[dynamic_scale.DynamicScale] = None

    @classmethod
    def init(cls, *args, **kwargs) -> Any:
        pass

    @classmethod
    def init_from_dict(cls, *args, **kwargs) -> TrainState:
        pass



from fortuna.calib_model.calib_mixin import WithCalibCheckpointingMixin
from fortuna.training.train_state_repository import TrainStateRepository


class CalibStateRepository(WithCalibCheckpointingMixin, TrainStateRepository):
    pass



from typing import (
    Any,
    Callable,
    Dict,
    Tuple,
    Union,
)

from flax import jax_utils
import jax
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from jax.tree_util import tree_map

from fortuna.data import TargetsLoader
from fortuna.output_calib_model.state import OutputCalibState
from fortuna.training.output_calibrator import (
    JittedMixin,
    MultiDeviceMixin,
    OutputCalibratorABC,
)
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
)


class ProbModelOutputCalibrator(OutputCalibratorABC):
    def training_loss_step(
        self,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: CalibParams,
        batch: Batch,
        outputs: Array,
        mutable: CalibMutable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        loss, aux = loss_fun(
            batch,
            n_data=n_data,
            return_aux=["outputs", "calib_mutable"],
            ensemble_outputs=outputs,
            calib_params=params,
            calib_mutable=mutable,
            rng=rng,
        )
        logging_kwargs = None
        return (
            loss,
            {
                "outputs": aux.get("outputs"),
                "mutable": aux.get("calib_mutable"),
                "logging_kwargs": logging_kwargs,
            },
        )

    def val_loss_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]:
        loss, aux = loss_fun(
            batch,
            n_data=n_data,
            return_aux=["outputs"],
            ensemble_outputs=outputs,
            calib_params=state.params,
            calib_mutable=state.mutable,
            rng=rng,
        )
        return loss, aux

    def __str__(self):
        return "calibration"


class ProbModelMultiDeviceMixin(MultiDeviceMixin):
    @staticmethod
    def _add_device_dim_to_outputs_loader(
        outputs_loader: TargetsLoader,
    ) -> TargetsLoader:
        def _reshape_batch(batch):
            n_devices = jax.local_device_count()
            if batch.shape[1] % n_devices != 0:
                raise ValueError(
                    f"The size of all output batches must be a multiple of {n_devices}, that is the number of "
                    f"available devices. However, a batch of outputs with shape {batch.shape[1]} was found. "
                    f"Please set an appropriate batch size."
                )
            shape = batch.shape
            return (
                batch.swapaxes(0, 1)
                .reshape(n_devices, shape[1] // n_devices, shape[0], shape[2])
                .swapaxes(1, 2)
            )

        class TargetsLoaderWrapper:
            def __init__(self, outputs_loader: TargetsLoader):
                self._outputs_loader = outputs_loader

            def __iter__(self):
                outputs_loader = map(
                    lambda batch: tree_map(_reshape_batch, batch), self._outputs_loader
                )
                outputs_loader = jax_utils.prefetch_to_device(outputs_loader, 2)
                yield from outputs_loader

        return (
            TargetsLoaderWrapper(outputs_loader)
            if outputs_loader is not None
            else outputs_loader
        )


class JittedProbModelOutputCalibrator(JittedMixin, ProbModelOutputCalibrator):
    pass


class MultiDeviceProbModelOutputCalibrator(
    ProbModelMultiDeviceMixin, ProbModelOutputCalibrator
):
    pass



import abc
from typing import (
    Dict,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from flax import linen as nn
from flax.core import FrozenDict
from flax.training.checkpoints import PyTree
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.typing import (
    InputData,
    Mutable,
    Params,
)
from fortuna.utils.random import WithRNG


class ModelManager(WithRNG, abc.ABC):
    """
    Abstract model manager class.
    It orchestrates the forward pass of the models in the probabilistic model.
    """

    def __init__(self, model: nn.Module, model_editor: Optional[nn.Module] = None):
        self.model = model
        self.model_editor = model_editor

    @abc.abstractmethod
    def apply(
        self,
        params: Params,
        inputs: InputData,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        """
        Apply the models' forward pass.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs : InputData
            Input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        train : bool
            Whether the method is called during training.
        rng: Optional[PRNGKeyArray]
            A random number generator.
            If not passed,
            this will be taken from the attributes of this class.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]
            The output of the model manager for each input. Mutable objects may also be returned.
        """
        pass

    @abc.abstractmethod
    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, Mapping]:
        """
        Initialize random parameters and mutable objects.

        Parameters
        ----------
        input_shape : Tuple
            The shape of the input variable.
        rng: Optional[PRNGKeyArray]
            A random number generator.
            If not passed,
            this will be taken from the attributes of this class.

        Returns
        -------
        Dict[str, FrozenDict]
            Initialized random parameters and mutable objects.
        """
        pass
